
from typing import Union as _Union
from typing import Callable as _Callable
from typing import List as _List

from .._network import Network
from .._networks import Networks
from .._population import Population
from .._population import Populations
from .._infections import Infections
from .._outputfiles import OutputFiles
from .._workspace import Workspace
from ._profiler import Profiler

__all__ = ["get_functions", "get_model_loop_functions",
           "get_initialise_functions", "get_finalise_functions",
           "get_summary_functions",
           "accepts_stage", "MetaFunction",
           "call_function_on_network"]


MetaFunction = _Callable[..., None]


def accepts_stage(func: MetaFunction) -> bool:
    """Return whether the passed function accepts the "stage" argument,
       meaning that it can do different things for different day stages

       Parameters
       ----------
       func: function
         The function to be queries

       Returns
       -------
       result: bool
         Whether or not the function accepts the "stage" argument
    """
    import inspect
    return "stage" in inspect.signature(func).parameters


def get_functions(stage: str,
                  network: _Union[Network, Networks],
                  population: Population,
                  infections: Infections,
                  output_dir: OutputFiles,
                  workspace: Workspace,
                  iterator: MetaFunction,
                  extractor: MetaFunction,
                  mixer: MetaFunction,
                  mover: MetaFunction,
                  rngs, nthreads, profiler: Profiler,
                  trajectory: Populations = None,
                  results=None) -> _List[MetaFunction]:
    """Return the functions that must be called for the specified
       stage of the day;

       * "initialise": model initialisation. Called once before the
                       whole model run is performed
       * "setup": day setup. Called once at the start of the day.
                  Should be used to import new seeds, move populations
                  between demographics, move infected individuals
                  through disease stages etc. There is no calculation
                  performed at this stage.
       * "foi": foi calculation. Called to recalculate the force of infection
                (foi) for each ward in the network (and subnetworks).
                This is calculated based on the populations in each state
                in each ward in each demographic
       * "infect": Called to advance the outbreak by calculating
                   the number of new infections
       * "analyse": Called at the end of the day to analyse
                    the data and extract the results
       * "finalise": Called at the end of the model run to finalise
                     any outputs or produce overall summary files
       * "summary": Called at the end of lots of model runs, to write
                    overview summary files. Only the extractor has
                    a summary stage

       Parameters
       ----------
       stage: str
         The stage of the day/model for which to get the functions
       network: Network or Networks
         The network(s) to be modelled
       population: Population
         The population experiencing the outbreak
       infections: Infections
         Space to record the infections through the day
       iterator: function
         Iterator used to obtain the function used to advance
         the outbreak
       extractor: function
         Extractor used to analyse the data and output results
       mixer: function
         Mixer used to mix and merge data between different demographics
       mover: function
         Mover used to move populations between demographics
       rngs: list[random number generate pointers]
         Pointers to the random number generators to use for each thread
       nthreads: int
         The number of threads to use to run the model
       profiler: Profiler
         The profiler used to profile the calculation

       Returns
       -------
       functions: List[MetaFunction]
         The list of all functions that should be called for this
         stage of the day
    """

    stages = ["initialise", "setup", "foi", "infect",
              "analyse", "finalise", "summary"]

    if stage not in stages:
        raise ValueError(
            f"Cannot recognise the stage {stage}. Available stages "
            f"are {stages}")

    kwargs = {"stage": stage,
              "network": network,
              "population": population,
              "infections": infections,
              "rngs": rngs,
              "nthreads": nthreads,
              "profiler": profiler,
              "trajectory": trajectory,
              "results": results}

    if stage == "summary":
        funcs = extractor(**kwargs)
    else:
        funcs = mover(**kwargs) + iterator(**kwargs) + \
            mixer(**kwargs) + extractor(**kwargs)

    return funcs


def get_model_loop_functions(**kwargs) -> _List[MetaFunction]:
    """Convenience function that returns all of the functions
       that should be called during the model loop
       (i.e. the "setup", "foi", "infect" and "analyse" stages)

       Parameters
       ----------
       network: Network or Networks
         The network(s) to be modelled
       population: Population
         The population experiencing the outbreak
       infections: Infections
         Space to record the infections through the day
       iterator: function
         Iterator used to obtain the function used to advance
         the outbreak
       extractor: function
         Extractor used to analyse the data and output results
       mixer: function
         Mixer used to mix and merge data between different demographics
       mover: function
         Mover used to move populations between demographics
       rngs: list[random number generate pointers]
         Pointers to the random number generators to use for each thread
       nthreads: int
         The number of threads to use to run the model
       profiler: Profiler
         The profiler used to profile the calculation

       Returns
       -------
       functions: List[MetaFunction]
         The list of all functions that should be called for this
         stage of the day
    """

    funcs = get_functions(stage="setup", **kwargs)
    funcs += get_functions(stage="foi", **kwargs)
    funcs += get_functions(stage="infect", **kwargs)
    funcs += get_functions(stage="analyse", **kwargs)

    return funcs


def get_initialise_functions(**kwargs) -> _List[MetaFunction]:
    """Convenience function that returns all of the functions
       that should be called during the initialisation step
       of the model (e.g. the "initialise" stage)

       Parameters
       ----------
       network: Network or Networks
         The network(s) to be modelled
       population: Population
         The population experiencing the outbreak
       infections: Infections
         Space to record the infections through the day
       iterator: function
         Iterator used to obtain the function used to advance
         the outbreak
       extractor: function
         Extractor used to analyse the data and output results
       mixer: function
         Mixer used to mix and merge data between different demographics
       mover: function
         Mover used to move populations between demographics
       rngs: list[random number generate pointers]
         Pointers to the random number generators to use for each thread
       nthreads: int
         The number of threads to use to run the model
       profiler: Profiler
         The profiler used to profile the calculation

       Returns
       -------
       functions: List[MetaFunction]
         The list of all functions that should be called for this
         stage of the day
    """
    return get_functions(stage="initialise", **kwargs)


def get_finalise_functions(trajectory: Populations,
                           **kwargs) -> _List[MetaFunction]:
    """Convenience function that returns all of the functions
       that should be called during the finalisation step
       of the model (e.g. the "finalise" stage)

       Parameters
       ----------
       network: Network or Networks
         The network(s) to be modelled
       population: Population
         The population experiencing the outbreak
       trajectory: Populations
         The trajectory of populations over time
       infections: Infections
         Space to record the infections through the day
       iterator: function
         Iterator used to obtain the function used to advance
         the outbreak
       extractor: function
         Extractor used to analyse the data and output results
       mixer: function
         Mixer used to mix and merge data between different demographics
       mover: function
         Mover used to move populations between demographics
       rngs: list[random number generate pointers]
         Pointers to the random number generators to use for each thread
       nthreads: int
         The number of threads to use to run the model
       profiler: Profiler
         The profiler used to profile the calculation

       Returns
       -------
       functions: List[MetaFunction]
         The list of all functions that should be called for this
         stage of the day
    """
    return get_functions(stage="finalise", **kwargs)


def get_summary_functions(network: _Union[Network, Networks],
                          results, output_dir: OutputFiles,
                          extractor: MetaFunction,
                          nthreads: int = 1, **kwargs
                          ):
    """Convenience function that returns all of the functions
       that should be called during the summary report
       stage of the simulation
    """
    kwargs["workspace"] = Workspace()
    kwargs["infections"] = Infections()
    kwargs["population"] = Population()
    kwargs["iterator"] = None
    kwargs["mixer"] = None
    kwargs["mover"] = None
    kwargs["rngs"] = None
    kwargs["profiler"] = None

    return get_functions(stage="summary", network=network,
                         output_dir=output_dir, results=results,
                         extractor=extractor, nthreads=nthreads, **kwargs)


def call_function_on_network(network: _Union[Network, Networks],
                             infections: Infections,
                             workspace: Workspace,
                             population: Population,
                             func: MetaFunction = None,
                             parallel: MetaFunction = None,
                             nthreads: int = 1,
                             switch_to_parallel: int = 2,
                             call_on_overall: bool = False,
                             **kwargs):
    """Call either 'func' or 'parallel' (depending on the
       number of threads, nthreads) on the passed Network,
       or on all demographic subnetworks

       Parameters
       ----------
       network: Network or Networks
         The network that is being modelled
       nthreads: int
         The number of threads to use to perform the work
       func: MetaFunction
         Function that performs the work in serial
       parallel: MetaFunction
         Function that performs the work in parallel
       switch_to_parallel: int
         Use the parallel function when nthreads is greater or equal
         to this value
    """
    if parallel is not None:
        if func is None or nthreads >= switch_to_parallel:
            kwargs["nthreads"] = nthreads
            func = parallel

    if isinstance(network, Networks):
        # call the function on all of the demographic sub-networks
        for i, subnet in enumerate(network.subnets):
            subinf = infections.subinfs[i]
            subwork = workspace.subspaces[i]
            subpop = population.subpops[i]

            func(network=subnet, infections=subinf,
                 population=subpop, workspace=subwork,
                 **kwargs)

        if call_on_overall:
            func(network=network.overall, infections=infections,
                 population=population, workspace=workspace,
                 **kwargs)
    else:
        func(network=network, infections=infections,
             population=population, workspace=workspace,
             **kwargs)




__all__ = ["import_graphics_modules",
           "save_summary_plots",
           "create_overview_plot",
           "create_average_plot",
           "create_demographics_plot"]


def import_graphics_modules(verbose=False):
    """Imports pandas and matplotlib in a safe way, giving a good
       error message if something goes wrong.

       Parameters
       ----------
       verbose: bool
         Whether or not to print to the screen to signal progress...

       Returns
       -------
       (pd, plt)
          The pandas (pd) and matplotlib.pyplot (plt) modules
    """
    try:
        if verbose:
            print("Importing graphics modules...")
        import pandas as pd
        import matplotlib.pyplot as plt
    except ImportError:
        print("You must have pandas and matplotlib installed to run "
              "metawards-plot")
        print("Install using either `pip install pandas` if you are using")
        print("pip, or 'conda install pandas' if you are using conda, ")
        print("or by running 'metawards-install --optional")
        raise ImportError("Cannot produce the plot as pandas and matplotlib "
                          "are not installed.")

    return (pd, plt)


def create_overview_plot(df, output_dir: str = None,
                         format: str = "jpg", dpi: int = 150,
                         align_axes: bool = True, verbose: bool = True):
    """Create a summary plot of the result.csv data held in the
       passed pandas dataframe. This returns the figure for you
       to save if desired (or just call ``plt.show()`` to show
       it in Jupyter)

       If the dataframe contains multiple fingerprints, then this
       will return a dictionary of figures, one for each fingerprint,
       indexed by fingerprint

       Parameters
       ----------
       df : Pandas Dataframe
         The pandas dataframe containing the data from results.csv.bz2
       output_dir: str
         The name of the directory in which to draw the graphs. If this
         is set then the graphs are written to files as they are generated
         and the filenames of the figures are returned. This is necessary
         when the number of graphs to draw is high and you don't want
         to waste too much memory
       format: str
         Format to save the figures in if output_dir is supplied
       dpi: int
         dpi (dots per inch) resolution to save the figures with if
         a bitmap format is used and output_dir is supplied
       align_axes: bool
         If true (default) then this will ensure that all of the plots
         for different fingerprints are put on the same axis scale
       verbose: bool
         Whether or not to print progress to the screen

       Returns
       -------
       fig
         The matplotlib figure containing the summary plot, or a
         dictionary of figures if there are multiple fingerprints,
         or the filename if output_dir was supplied, or a dictionary
         of multiple filenames indexed by fingerprint
    """
    _, plt = import_graphics_modules()

    try:
        fingerprints = df["fingerprint"].unique()
        repeat = "repeat"
    except Exception:
        # no fingerprints
        fingerprints = [None]
        repeat = "demographic"

    try:
        import PIL        # noqa - disable unused warning
    except ImportError:
        if format == "jpg":
            print(
                "WARNING: Missing 'pillow' package, defaulting to PNG format.")
            format = "png"

    figs = {}

    min_date = None
    max_date = None
    max_y = {}
    min_y = {}

    columns = ["E", "I", "IW", "R"]

    nfigs = len(fingerprints)

    if len(fingerprints) > 1 and align_axes:
        for fingerprint in fingerprints:
            df2 = df[df["fingerprint"] == fingerprint]

            for column in columns:
                min_d = df2["day"].min()
                max_d = df2["day"].max()
                min_val = df2[column].min()
                max_val = df2[column].max()

                if min_date is None:
                    min_date = min_d
                    max_date = max_d
                else:
                    if min_d < min_date:
                        min_date = min_d
                    if max_d > max_date:
                        max_date = max_d

                if column not in min_y:
                    min_y[column] = min_val
                    max_y[column] = max_val
                else:
                    if min_val < min_y[column]:
                        min_y[column] = min_val
                    if max_val > max_y[column]:
                        max_y[column] = max_val

    for fingerprint in fingerprints:
        if fingerprint is None:
            df2 = df
        else:
            df2 = df[df["fingerprint"] == fingerprint]

        fig, axes = plt.subplots(nrows=2, ncols=2, figsize=(10, 10))

        i = 0
        j = 0

        for column in columns:
            ax = df2.pivot(index="date", columns=repeat,
                           values=column).plot.line(ax=axes[i][j])
            ax.tick_params('x', labelrotation=90)
            ax.get_legend().remove()
            ax.set_ylabel("Population")

            if len(fingerprints) > 1 and align_axes:
                ax.set_xlim(min_date, max_date)
                ax.set_ylim(min_y[column], 1.1*max_y[column])

            if len(fingerprints) > 1:
                from metawards import VariableSet
                fvals, _rpt = VariableSet.extract_values(fingerprint)
                ax.set_title(f"{fvals} : {column}")
            else:
                ax.set_title(column)

            j += 1
            if j == 2:
                j = 0
                i += 1

        fig.tight_layout(pad=1)

        if output_dir:
            import os

            if nfigs == 1:
                filename = os.path.join(output_dir, f"overview.{format}")
            else:
                filename = os.path.join(output_dir,
                                        f"overview_{fingerprint}.{format}")

            if verbose:
                print(f"Saving figure {filename}")

            fig.savefig(filename, dpi=dpi)
            plt.close()
            fig = None
            figs[fingerprint] = filename
        else:
            if verbose:
                print(f"Created the figure for {fingerprint}")

            figs[fingerprint] = fig

    if len(figs) == 0:
        return None
    elif len(figs) == 1:
        return figs[list(figs.keys())[0]]
    else:
        return figs


def create_average_plot(df, output_dir: str = None, format: str = "jpg",
                        dpi: int = 150, align_axes: bool = True,
                        verbose: bool = True):
    """Create an average plot of the result.csv data held in the
       passed pandas dataframe. This returns the figure for you
       to save if desired (or just call ``plt.show()`` to show
       it in Jupyter)

       Note that this won't do anything unless there are multiple
       repeats of the model run in the output. In that case, it
       will return None

       If the dataframe contains multiple fingerprints, then this
       will return a dictionary of figures, one for each fingerprint,
       indexed by fingerprint

       Parameters
       ----------
       df : Pandas Dataframe
         The pandas dataframe containing the data from results.csv.bz2
       output_dir: str
         The name of the directory in which to draw the graphs. If this
         is set then the graphs are written to files as they are generated
         and the filenames of the figures are returned. This is necessary
         when the number of graphs to draw is high and you don't want
         to waste too much memory
       format: str
         Format to save the figures in if output_dir is supplied
       dpi: int
         dpi (dots per inch) resolution to save the figures with if
         a bitmap format is used and output_dir is supplied
       align_axes: bool
         If true (default) then this will ensure that all of the plots
         for different fingerprints are put on the same axis scale
       verbose: bool
         Whether or not to print progress to the screen

       Returns
       -------
       fig
         The matplotlib figure containing the average plot, or a
         dictionary of figures if there are multiple fingerprints,
         or the filename if output_dir was supplied, or a dictionary
         of multiple filenames indexed by fingerprint
    """
    try:
        import PIL        # noqa - disable unused warning
    except ImportError:
        if format == "jpg":
            print(
                "WARNING: Missing 'pillow' package, defaulting to PNG format.")
            format = "png"

    fingerprints = df["fingerprint"].unique()

    figs = {}

    nfigs = len(fingerprints)

    for fingerprint in fingerprints:
        df2 = df[df["fingerprint"] == fingerprint]

        nrepeats = len(df2["repeat"].unique())

        if nrepeats > 1:
            _, plt = import_graphics_modules()

            fig, axes = plt.subplots(nrows=2, ncols=2, figsize=(10, 10))

            mean_average = df2.groupby("date").mean()
            stddev = df2.groupby("date").std()

            i = 0
            j = 0

            for column in ["E", "I", "IW", "R"]:
                ax = mean_average.plot.line(y=column, yerr=stddev[column],
                                            ax=axes[i][j])
                ax.tick_params('x', labelrotation=90)
                ax.get_legend().remove()
                ax.set_title(column)
                ax.set_ylabel("Population")

                j += 1
                if j == 2:
                    j = 0
                    i += 1

            fig.tight_layout(pad=1)

            if output_dir:
                import os

                if nfigs == 1:
                    filename = os.path.join(output_dir, f"average.{format}")
                else:
                    filename = os.path.join(output_dir,
                                            f"average_{fingerprint}.{format}")

                if verbose:
                    print(f"Saving figure {filename}")

                fig.savefig(filename, dpi=dpi)
                fig = None
                plt.close()
                figs[fingerprint] = filename
            else:
                if verbose:
                    print(f"Created the figure for {fingerprint}")

                figs[fingerprint] = fig

    if len(figs) == 0:
        return None
    elif len(figs) == 1:
        return figs[list(figs.keys())[0]]
    else:
        return figs


def get_color(name):
    """Return a good color for the passed name"""
    name = str(name).strip().lower()

    if name == "overall":
        return "black"

    elif name in ["red", "blue", "green", "orange", "yellow", "black",
                  "white", "gray", "pink"]:
        return name

    else:
        # return a random colour
        import random
        rgb = (random.random(), random.random(), random.random())
        return rgb


def create_demographics_plot(df, output_dir: str = None,
                             format: str = "jpg", dpi: int = 150,
                             verbose: bool = True):
    """Create a demographics plot of the trajectory.csv data held in the
       passed pandas dataframe. This returns the figure for you
       to save if desired (or just call ``plt.show()`` to show
       it in Jupyter)

       Parameters
       ----------
       df : Pandas Dataframe
         The pandas dataframe containing the data from trajectory.csv.bz2
       output_dir: str
         The name of the directory in which to draw the graphs. If this
         is set then the graphs are written to files as they are generated
         and the filenames of the figures are returned. This is necessary
         when the number of graphs to draw is high and you don't want
         to waste too much memory
       format: str
         Format to save the figures in if output_dir is supplied
       dpi: int
         dpi (dots per inch) resolution to save the figures with if
         a bitmap format is used and output_dir is supplied
       align_axes: bool
         If true (default) then this will ensure that all of the plots
         for different fingerprints are put on the same axis scale
       verbose: bool
         Whether or not to print progress to the screen

       Returns
       -------
       fig
         The matplotlib figure containing the demographics plot, or
         the filename if output_dir was supplied
    """
    try:
        import PIL        # noqa - disable unused warning
    except ImportError:
        if format == "jpg":
            print(
                "WARNING: Missing 'pillow' package, defaulting to PNG format.")
            format = "png"

    _, plt = import_graphics_modules()

    columns = ["E", "I", "IW", "R"]

    fig, axes = plt.subplots(nrows=2, ncols=2, figsize=(10, 10))

    # see if any of these give colour names
    colors = []

    demographics = df.pivot(index="date", columns="demographic",
                            values="day").columns

    for demographic in demographics:
        colors.append(get_color(demographic))

    i = 0
    j = 0

    for column in columns:
        ax = df.pivot(index="date", columns="demographic",
                      values=column).plot.line(ax=axes[i][j],
                                               color=colors)
        ax.tick_params('x', labelrotation=90)
        ax.set_ylabel("Population")
        ax.set_title(column)

        j += 1
        if j == 2:
            j = 0
            i += 1

    fig.tight_layout(pad=1)

    if output_dir:
        import os

        filename = os.path.join(output_dir, f"demographics.{format}")

        if verbose:
            print(f"Saving figure {filename}")

        fig.savefig(filename, dpi=dpi)
        plt.close()
        fig = filename

    return fig


def save_summary_plots(results: str, output_dir: str = None,
                       format: str = "jpg", dpi: int = 150,
                       align_axes: bool = True,
                       verbose=False):
    """Create summary plots of the data contained in the passed
       'results.csv.bz2' file that was produced by metawards
       and save them to disk.

       Parameters
       ----------
       results: str
         The full path to the file containing the results. This
         **must** have been created by ``metawards``
       output_dir: str
         Path to the directory in which you want to place the graphs.
         This defaults to the same directory that contains 'results'
       format: str
         The format to use to save the graphs. This defaults to 'pdf'
       dpi: int
         The dots-per-inch to use when saving bitmap graphics (e.g.
         png, jpg etc)
       align_axes: bool
         Whether or not to plot all graphs in a set on the same axes
       verbose: bool
         Whether or not to print progress to the screen

       Returns
       -------
       filenames: List(str)
         Full file paths of all of the files written by this function
    """
    try:
        import PIL        # noqa - disable unused warning
    except ImportError:
        if format == "jpg":
            print(
                "WARNING: Missing 'pillow' package, defaulting to PNG format.")
            format = "png"

    pd, _ = import_graphics_modules(verbose=verbose)
    import os

    if verbose:
        print(f"Reading data from {results}...")

    df = pd.read_csv(results)

    if output_dir is None:
        output_dir = os.path.dirname(results)

        if output_dir is None or len(output_dir) == 0:
            output_dir = "."

    if format is None:
        format = "pdf"

    filenames = []

    # is this an output from multiple runs?
    try:
        df["fingerprint"]
        has_fingerprint = True
    except Exception:
        has_fingerprint = False
        pass

    # does this have demographic data?
    try:
        df["demographic"]
        has_demographics = True
    except Exception:
        has_demographics = False

    if has_fingerprint:
        if verbose:
            print(f"Creating overview plot(s)...")

        figs = create_overview_plot(df, output_dir=output_dir,
                                    format=format, dpi=dpi,
                                    align_axes=align_axes)

        if isinstance(figs, dict):
            filenames += list(figs.values())
        elif figs is not None:
            filenames.append(figs)

        if verbose:
            print(f"Creating average plot(s)...")

        figs = create_average_plot(df, output_dir=output_dir,
                                   format=format, dpi=dpi,
                                   align_axes=align_axes)

        if isinstance(figs, dict):
            filenames += list(figs.values())
        elif figs is not None:
            filenames.append(figs)

    if has_demographics:
        fig = create_demographics_plot(df, output_dir=output_dir,
                                       format=format, dpi=dpi)

        if fig is not None:
            filenames.append(fig)

    return filenames




from typing import Union as _Union

from .._network import Network
from .._networks import Networks
from .._population import Population
from .._infections import Infections

__all__ = ["setup_seed_specified_ward",
           "setup_seed_all_wards",
           "setup_seed_wards"]


def setup_seed_specified_ward(network: _Union[Network, Networks],
                              infections: Infections,
                              **kwargs):
    """Setup function that sets up the Network by seeding the infection
       at the ward specified by looking up the ward index
       from network.to_seed using index params.ward_seed_index

       Parameters
       ----------
       network: Network or Networks
         The network to be seeded
       infections: Infections
         Space to hold the 'work' infections
       kwargs
         Arguments that are not used by this setup function
    """
    if isinstance(network, Networks):
        # only seed the overall ward
        raise NotImplementedError(
            "Seeding a specific ward for a multi-demographic Network is "
            "not yet supported. If you need this please raise an issue "
            "at https://github.com/metawards/MetaWards/issues and we will "
            "work as quickly as we can to implement it.")

    wards = network.nodes
    links = network.links
    params = network.params

    play_infections = infections.play
    infections = infections.work

    seed_index = int(params.ward_seed_index)
    seed = network.to_seed[seed_index]

    from ._console import Console

    Console.print(f"Setup by seeding the ward at index {seed}")

    j = 0

    while (links.ito[j] != seed) or (links.ifrom[j] != seed):
        j += 1

    if links.suscept[j] < params.initial_inf:
        wards.play_suscept[seed] -= params.initial_inf
        play_infections[0][seed] += params.initial_inf

    infections[0][j] = params.initial_inf
    links.suscept[j] -= params.initial_inf


def setup_seed_all_wards(network: _Union[Network, Networks],
                         population: Population,
                         infections: Infections,
                         **kwargs):
    """Seed the wards with an initial set of infections, assuming
       an 'expected' number of infected people out of a population
       of 'population', based on the number of daily imports held
       in params.daily_imports
    """
    if isinstance(network, Networks):
        if network.overall.params.daily_imports == 0:
            return

        raise NotImplementedError(
            "Daily seeding of a multi-demographic network is not yet "
            "supported. If you need this, please raise an issue on "
            "https://github.com/metawards/MetaWards/issues and we will "
            "work as quickly as we can to implement it.")

    wards = network.nodes
    params = network.params

    if params.daily_imports == 0:
        return

    play_infections = infections.play
    infections = infections.work

    from ._console import Console
    Console.print(f"Setup by seeding all wards")

    frac = float(params.daily_imports) / float(population.initial)

    for i in range(0, network.nnodes+1):  # 1-index but also count at 0?
        temp = wards.denominator_n[i] + wards.denominator_p[i]
        to_seed = int(frac*temp + 0.5)
        wards.play_suscept[i] -= to_seed
        play_infections[0][i] += to_seed


def setup_seed_wards(network: Network,
                     **kwargs):
    """Seed the wards with an initial infection. If a specific ward
       has been highlighted via `params.ward_seed_index`, then this
       will be seeded using the function `setup_seed_specified_ward`.
       Otherwise, all wards will be seeded via `setup_seed_all_wards`.

       Parameters
       ----------
       network: Network
         The network over which the model will run
       kwargs
         Arguments not needed for this function
    """

    params = network.params

    try:
        seed_index = int(params.ward_seed_index)
    except Exception:
        seed_index = None

    if seed_index is None:
        setup_seed_all_wards(network=network,
                             **kwargs)
    else:
        setup_seed_specified_ward(network=network,
                                  **kwargs)




from dataclasses import dataclass as _dataclass
from typing import Union as _Union

from ._network import Network
from ._networks import Networks

__all__ = ["Infections"]


@_dataclass
class Infections:
    """This class holds the arrays that record the infections as they
       are occuring during the outbreak
    """

    #: The infections caused by fixed (work) movements. This is a list
    #: of int arrays, size work[N_INF_CLASSES][nlinks+1]
    work = None

    #: The infections caused by random (play) movements. This is a list
    #: of int arrays, size play[N_INF_CLASSES][nnodes+1]
    play = None

    #: The infections for the multi-demographic subnets
    subinfs = None

    #: The index in the overall network's work matrix of the ith
    #: index in this subnetworks work matrix. If this is None then
    #: both this subnetwork has the same work matrix as the overall
    #: network
    _work_index = None

    @property
    def N_INF_CLASSES(self) -> int:
        """The total number of stages in the disease"""
        if self.work is not None:
            return len(self.work)
        else:
            return 0

    @property
    def nnodes(self) -> int:
        """The total number of nodes (wards)"""
        if self.play is not None and len(self.play) > 0:
            return len(self.play[0]) - 1    # 1-indexed
        else:
            return 0

    @property
    def nlinks(self) -> int:
        """Return the number of work links"""
        if self.work is not None and len(self.work) > 0:
            return len(self.work[0]) - 1    # 1-indexed
        else:
            return 0

    @property
    def nsubnets(self) -> int:
        """Return the number of demographic subnetworks"""
        if self.subinfs is not None:
            return len(self.subinfs)
        else:
            return 0

    @staticmethod
    def build(network: _Union[Network, Networks] = None):
        """Construct and return the Infections object that will track
           infections during a model run on the passed Network (or Networks)

           Parameters
           ----------
           network: Network or Networks
             The network or networks that will be run

           Returns
           -------
           infections: Infections
             The space for the work and play infections for the network
             (including space for all of the demographics)
        """
        from .utils import initialise_infections, initialise_play_infections

        if isinstance(network, Network):
            inf = Infections()
            inf.work = initialise_infections(network=network)
            inf.play = initialise_play_infections(network=network)

            inf._ifrom = network.links.ifrom
            inf._ito = network.links.ito

            return inf

        elif isinstance(network, Networks):
            inf = Infections.build(network.overall)

            subinfs = []

            for subnet in network.subnets:
                subinfs.append(Infections.build(subnet))

            inf.subinfs = subinfs

            return inf

    def has_different_work_matrix(self):
        """Return whether or not the sub-network work matrix
           is different to that of the overall network
        """
        return self._work_index is not None

    def get_work_index(self):
        """Return the mapping from the index in this sub-networks work
           matrix to the mapping in the overall network's work matrix
        """
        if self.has_different_work_matrix():
            # remember this is 1-indexed, so work_index[1] is the first
            # value
            return self._work_index
        else:
            return range(1, self.nlinks + 1)

    def aggregate(self, profiler=None, nthreads: int = 1) -> None:
        """Aggregate all of the infection data from the demographic
           sub-networks

           Parameters
           ----------
           network: Network
               Network that was used to initialise these infections
           profiler : Profiler, optional
               Profiler used to profile the calculation, by default None
           nthreads : int, optional
               Number of threads to use, by default 1
        """
        from .utils._aggregate import aggregate_infections
        aggregate_infections(infections=self, profiler=profiler,
                             nthreads=nthreads)

    def clear(self, nthreads: int = 1):
        """Clear all of the infections (resets all to zero)

           Parameters
           ----------
           nthreads: int
             Optionally parallelise this reset by specifying the number
             of threads to use
        """
        from .utils import clear_all_infections
        clear_all_infections(infections=self.work,
                             play_infections=self.play,
                             nthreads=nthreads)

        if self.subinfs is not None:
            for subinf in self.subinfs:
                subinf.clear(nthreads=nthreads)




from typing import Union as _Union
from typing import List as _List

cimport cython
from cython.parallel import parallel, prange
from libc.stdint cimport uintptr_t

from .._networks import Networks
from .._infections import Infections
from .._demographics import DemographicID, DemographicIDs

from ..utils._console import Console
from ..utils._profiler import Profiler
from ..utils._get_array_ptr cimport get_int_array_ptr, get_double_array_ptr

from ..utils._ran_binomial cimport _ran_binomial, \
                                   _get_binomial_ptr, binomial_rng

__all__ = ["go_to", "go_to_serial", "go_to_parallel"]


def go_to_parallel(go_from: _Union[DemographicID, DemographicIDs],
                   go_to: DemographicID,
                   network: Networks,
                   infections: Infections,
                   profiler: Profiler,
                   rngs,
                   nthreads: int,
                   fraction: float = 1.0,
                   **kwargs) -> None:
    """This go function will move individuals from the "from"
       demographic(s) to the "to" demographic. This can move
       a subset of individuals if 'fraction' is less than 1, e.g.
       0.5 would move 50% of individuals (chosen using
       a random binomial distribution)

       Parameters
       ----------
       from: DemographicID or DemographicIDs
         ID (name or index) or IDs of the demographics to scan for
         infected individuals
       to: DemographicID
         ID (name or index) of the demographic to send infected
         individuals to
       network: Networks
         The networks to be modelled. This must contain all of the
         demographics that are needed for this go function
       fraction: float
         The fraction of individuals to move, e.g. 0.75 would move
         75% of the individuals
       rngs:
         Thread-safe random number generators used to choose the
         individuals to move
    """

    if fraction < 0 or fraction > 1:
        raise ValueError(
            f"The move fraction {fraction} should be from 0 to 1")

    if fraction == 0:
        # nothing to do
        return

    # make sure that all of the needed demographics exist, and
    # convert them into a canonical form (indexes, list[indexes])
    if not isinstance(go_from, list):
        go_from = [go_from]

    subnets = network.subnets
    demographics = network.demographics
    subinfs = infections.subinfs

    go_from = [demographics.get_index(x) for x in go_from]

    go_to = demographics.get_index(go_to)

    if go_to in go_from:
        raise ValueError(
            f"You cannot move to {go_to} as it is also in {go_from}")

    to_subnet = subnets[go_to]
    to_subinf = subinfs[go_to]

    cdef int N_INF_CLASSES = infections.N_INF_CLASSES

    cdef double frac = fraction

    cdef int nnodes_plus_one = 0
    cdef int nlinks_plus_one = 0

    cdef int * work_infections_i
    cdef int * play_infections_i

    cdef int * to_work_infections_i
    cdef int * to_play_infections_i

    cdef double * links_suscept
    cdef double * links_weight

    cdef double * nodes_play_suscept
    cdef double * nodes_save_play_suscept

    cdef double * to_links_suscept = get_double_array_ptr(
                                            to_subnet.links.suscept)
    cdef double * to_links_weight = get_double_array_ptr(
                                            to_subnet.links.weight)

    cdef double * to_nodes_play_suscept = get_double_array_ptr(
                                            to_subnet.nodes.play_suscept)
    cdef double * to_nodes_save_play_suscept = get_double_array_ptr(
                                            to_subnet.nodes.save_play_suscept)

    cdef int nsubnets = len(subnets)

    # get the random number generator
    cdef uintptr_t [::1] rngs_view = rngs
    cdef binomial_rng* rng   # pointer to parallel rng

    cdef int thread_id
    cdef int num_threads = nthreads

    cdef int ii = 0
    cdef int i = 0
    cdef int j = 0

    cdef int to_move = 0
    cdef double to_move_d = 0.0

    for ii in go_from:
        subnet = subnets[ii]
        subinf = subinfs[ii]
        nnodes_plus_one = subinf.nnodes + 1
        nlinks_plus_one = subinf.nlinks + 1

        # move the susceptibles
        links_suscept = get_double_array_ptr(subnet.links.suscept)
        links_weight = get_double_array_ptr(subnet.links.weight)
        nodes_play_suscept = get_double_array_ptr(subnet.nodes.play_suscept)
        nodes_save_play_suscept = get_double_array_ptr(
                                                subnet.nodes.save_play_suscept)

        with nogil, parallel(num_threads=num_threads):
            if frac == 1.0:
                for j in prange(1, nlinks_plus_one, schedule="static"):
                    to_move_d = links_suscept[j]

                    if to_move_d > 0:
                        to_links_suscept[j] = to_links_suscept[j] + \
                                              to_move_d
                        links_suscept[j] = links_suscept[j] - to_move_d

                        to_links_weight[j] = to_links_suscept[j]
                        links_weight[j] = links_suscept[j]

                for j in prange(1, nnodes_plus_one, schedule="static"):
                    to_move_d = nodes_play_suscept[j]

                    if to_move_d > 0:
                        to_nodes_play_suscept[j] = to_nodes_play_suscept[j] + \
                                                   to_move_d
                        nodes_play_suscept[j] = nodes_play_suscept[j] - \
                                                to_move_d

                        to_nodes_save_play_suscept[j] = \
                                                    to_nodes_play_suscept[j]
                        nodes_save_play_suscept[j] = nodes_play_suscept[j]
            else:
                thread_id = cython.parallel.threadid()
                rng = _get_binomial_ptr(rngs_view[thread_id])

                for j in prange(1, nlinks_plus_one, schedule="static"):
                    to_move = _ran_binomial(rng, frac,
                                            <int>(links_suscept[j]))

                    if to_move > 0:
                        to_links_suscept[j] = to_links_suscept[j] + \
                                              to_move
                        links_suscept[j] = links_suscept[j] - to_move

                        to_links_weight[j] = to_links_suscept[j]
                        links_weight[j] = links_suscept[j]

                for j in prange(1, nnodes_plus_one, schedule="static"):
                    to_move = _ran_binomial(rng, frac,
                                            <int>(nodes_play_suscept[j]))

                    if to_move > 0:
                        to_nodes_play_suscept[j] = to_nodes_play_suscept[j] + \
                                                   to_move
                        nodes_play_suscept[j] = nodes_play_suscept[j] - \
                                                to_move

                        to_nodes_save_play_suscept[j] = \
                                                    to_nodes_play_suscept[j]
                        nodes_save_play_suscept[j] = nodes_play_suscept[j]

        # move infected / recovered individuals
        for i in range(0, N_INF_CLASSES):
            work_infections_i = get_int_array_ptr(subinf.work[i])
            play_infections_i = get_int_array_ptr(subinf.play[i])

            to_work_infections_i = get_int_array_ptr(to_subinf.work[i])
            to_play_infections_i = get_int_array_ptr(to_subinf.play[i])

            with nogil, parallel(num_threads=num_threads):
                if frac == 1.0:
                    for j in prange(1, nlinks_plus_one, schedule="static"):
                        to_move = work_infections_i[j]

                        if to_move > 0:
                            to_work_infections_i[j] = \
                                                to_work_infections_i[j] + \
                                                to_move
                            work_infections_i[j] = \
                                                work_infections_i[j] - \
                                                to_move

                    for j in prange(1, nnodes_plus_one, schedule="static"):
                        to_move = play_infections_i[j]

                        if to_move > 0:
                            to_play_infections_i[j] = \
                                                to_play_infections_i[j] + \
                                                to_move

                            play_infections_i[j] = \
                                                play_infections_i[j] - \
                                                to_move
                else:
                    # only move a fraction of the population
                    thread_id = cython.parallel.threadid()
                    rng = _get_binomial_ptr(rngs_view[thread_id])

                    for j in prange(1, nlinks_plus_one, schedule="static"):
                        to_move = _ran_binomial(rng, frac,
                                                <int>(work_infections_i[j]))

                        if to_move > 0:
                            to_work_infections_i[j] = \
                                                to_work_infections_i[j] + \
                                                to_move
                            work_infections_i[j] = \
                                                work_infections_i[j] - \
                                                to_move

                    for j in prange(1, nnodes_plus_one, schedule="static"):
                        to_move = _ran_binomial(rng, frac,
                                                <int>(play_infections_i[j]))

                        if to_move > 0:
                            to_play_infections_i[j] = \
                                                to_play_infections_i[j] + \
                                                to_move

                            play_infections_i[j] = \
                                                play_infections_i[j] - \
                                                to_move


def go_to_serial(**kwargs):
    go_to_parallel(nthreads=1, **kwargs)


def go_to(nthreads: int = 1, **kwargs) -> None:
    """This go function will move individuals from the "from"
       demographic(s) to the "to" demographic. This can move
       a subset of individuals is 'fraction' is less than 1, e.g.
       0.5 would move 50% of individuals.

       Parameters
       ----------
       from: DemographicID or DemographicIDs
         ID (name or index) or IDs of the demographics to scan for
         infected individuals
       to: DemographicID
         ID (name or index) of the demographic to send infected
         individuals to
       network: Networks
         The networks to be modelled. This must contain all of the
         demographics that are needed for this go function
       fraction: float
         The fraction of individuals to move, e.g. 0.75 would move
         75% of the individuals
    """

    if nthreads > 1:
        go_to_parallel(nthreads=nthreads, **kwargs)
    else:
        go_to_serial(**kwargs)




from .._network import Network

from ._profiler import Profiler

__all__ = ["get_min_max_distances"]


def get_min_max_distances(network: Network, nthreads: int = 1,
                          profiler: Profiler = None):
    """Return the minimum and maximum distances recorded in the network"""
    links = network.links

    cdef double mindist = -1.0
    cdef double maxdist = -1.0

    cdef int i = 0
    cdef double dist = 0.0
    cdef double [::1] links_distance = links.distance

    for i in range(1, network.nlinks+1):
        dist = links_distance[i]

        if dist:
            if mindist < 0.0:
                mindist = dist
                maxdist = dist
            elif dist > maxdist:
                maxdist = dist
            elif dist < mindist:
                mindist = dist

    from ._console import Console
    Console.print(f"maxdist {maxdist} mindist {mindist}")

    return (mindist, maxdist)




from typing import Union as _Union

from .._network import Network
from .._networks import Networks
from .._population import Populations
from .._outputfiles import OutputFiles

__all__ = ["output_trajectory"]


def output_trajectory(network: _Union[Network, Networks],
                      output_dir: OutputFiles,
                      trajectory: Populations,
                      **kwargs) -> None:
    """Call in the "finalise" stage to output the
       population trajectory to the 'trajectory.csv' file
    """

    RESULTS = output_dir.open("trajectory.csv")

    has_date = trajectory[0].date

    if has_date:
        datestring = "date,"
    else:
        datestring = ""

    RESULTS.write(f"day,{datestring}demographic,S,E,I,R,IW\n")

    for i, pop in enumerate(trajectory):
        if pop.date:
            d = pop.date.isoformat() + ","
        else:
            d = ""

        RESULTS.write(f"{pop.day},{d}overall,{pop.susceptibles},"
                      f"{pop.latent},{pop.total},"
                      f"{pop.recovereds},{pop.n_inf_wards}\n")

        if isinstance(network, Networks):
            for i, demographic in enumerate(network.demographics):
                subpop = pop.subpops[i]
                name = demographic.name

                if name is None or len(name) == 0:
                    name = str(i)

                RESULTS.write(f"{subpop.day},{d}{name},{subpop.susceptibles},"
                              f"{subpop.latent},{subpop.total},"
                              f"{subpop.recovereds},{subpop.n_inf_wards}\n")




from cpython cimport array

import array   # timing shows quicker for random access
               # than numpy

__all__ = ["create_int_array", "create_double_array",
           "create_string_array", "resize_array"]


cdef array.array _int_array_template = array.array('i', [])
cdef array.array _dbl_array_template = array.array('d', [])


def create_string_array(size: int, default: str=None):
    """Create an array of python strings of size 'size', optionally
       initialised with 'default'
    """
    return size * [default]   # this is a list of strings


def resize_array(a, size: int, default: any = None):
    """Resize the passed array to size 'size', adding 'default'
       if this will grow the array
    """
    if size == 0:
        from ._console import Console
        Console.warning("RESIZE TO ZERO")
        return a[0:0]
    elif size == len(a):
        return a
    elif size < len(a):
        return a[0:size]
    elif len(a) == 0:
        typecode = None

        try:
            typecode = a.typecode
        except:
            typecode = None

        if typecode is None:
            return create_string_array(size, default)
        elif typecode == "i":
            return create_int_array(size, default)
        elif typecode == "d":
            return create_double_array(size, default)
        else:
            raise AssertionError(f"Unrecognised array typecode {typecode}")

    elif isinstance(a[0], str):
        a = a + ((len(a) - size) * [default])
        assert(len(a) == size)
        return a
    else:
        try:
            typecode = a.typecode
        except:
            typecode = None

        d = size - len(a)

        if typecode is None:
            add = create_string_array(d, default)
        elif typecode == "i":
            add = create_int_array(d, default)
        elif typecode == "d":
            add = create_double_array(d, default)
        else:
            raise AssertionError(f"Unrecognised array typecode {typecode}")

        new_array = a + add
        return a


def create_double_array(size: int, default: float=None):
    """Create a new array.array of the specified size. If
       default is set then all values will be initialised to
       'default'
    """
    cdef int s = size
    cdef array.array dbl_array
    dbl_array = array.clone(_dbl_array_template, size, zero=False)

    cdef int i
    cdef double d

    cdef double [::1] view = dbl_array

    if default is not None:
        d = default
        for i in range(0,s):
            view[i] = d

    return dbl_array


def create_int_array(size: int, default: int=None):
    """Create a new array.array of the specified size. If
       default is set then all values will be initialised to
       'default'
    """
    cdef int s = size
    cdef array.array int_array
    int_array = array.clone(_int_array_template, size, zero=False)

    cdef int i, d

    cdef int [::1] view = int_array

    if default is not None:
        d = default
        for i in range(0,s):
            view[i] = d

    return int_array




__all__ = ["iterate_working_week"]

from ._iterate_weekday import iterate_weekday
from ._iterate_weekend import iterate_weekend

from .._population import Population


def iterate_working_week(population: Population, **kwargs):
    """This returns the default list of 'advance_XXX' functions that
       are called in sequence for each iteration of the model run.
       This iterator understands the concept of a traditional working week,
       namely Monday-Friday is a work day, while Saturday and
       Sunday are weekends

       Parameters
       ----------
       population: Population
         The population experiencing the outbreak. This includes
         information about the day and date of the outbreak

       Returns
       -------
       funcs: List[function]
         The list of functions that ```iterate``` will call in sequence
    """

    kwargs["population"] = population

    # is this a weekday or a weekend?
    if population.date is None:
        # have to guess from the day - assume day zero was a Monday
        day = population.day % 7
        # 0-4 is a weekday, 5 and 6 are weekend
        is_weekend = (day >= 5)
    else:
        day = population.date.weekday()
        # 0-4 is a weekday, 5 and 6 are weekend
        is_weekend = (day >= 5)

    if is_weekend:
        return iterate_weekend(**kwargs)
    else:
        return iterate_weekday(**kwargs)




__all__ = ["mix_none"]


def mix_none(**kwargs):
    """This mixer will perform no mixing. The result is that
       the demographics won't interact with one another and
       each disease outbreak will be completely separate.
    """
    return []



#!/bin/env/python3
#cython: linetrace=False
# MUST ALWAYS DISABLE AS WAY TOO SLOW FOR ITERATE

cimport cython
from cython.parallel import parallel, prange

from .._infections import Infections

from ._get_array_ptr cimport get_int_array_ptr

__all__ = ["clear_all_infections"]


def clear_all_infections(infections, play_infections,
                         nthreads: int=1):
    """Clears all infections associated with a model run

       Parameters
       ----------
       infections
         Space that is used to hold all of the 'work' infections
       play_infections
         Space that is used to hold all of the 'play' infections
       nthreads: int
         Number of threads to use to clear the arrays
    """

    cdef int i = 0
    cdef int j = 0
    cdef int n = len(infections)

    if n <= 0:
        return

    cdef int ninf = len(infections[0])
    cdef int nplay = len(play_infections[0])

    cdef int * a
    cdef int * b
    cdef int num_threads = nthreads

    for i in range(0, n):
        a = get_int_array_ptr(infections[i])
        b = get_int_array_ptr(play_infections[i])

        with nogil, parallel(num_threads=num_threads):
            for j in prange(0, ninf, schedule="static"):
                a[j] = 0

            for j in prange(0, nplay, schedule="static"):
                b[j] = 0



#!/bin/env/python3
#cython: linetrace=False
# MUST ALWAYS DISABLE AS WAY TOO SLOW FOR ITERATE

cimport cython
from cython.parallel import parallel, prange
cimport openmp

from libc.stdlib cimport calloc, free
from libc.stdint cimport uintptr_t
from libc.math cimport cos, pi


from .._network import Network
from .._population import Population
from .._infections import Infections

from ..utils._profiler import Profiler
from ..utils._get_functions import call_function_on_network

from ..utils._ran_binomial cimport _ran_binomial, \
                                   _get_binomial_ptr, binomial_rng

from ..utils._get_array_ptr cimport get_int_array_ptr, get_double_array_ptr

__all__ = ["advance_foi", "advance_foi_omp", "advance_foi_serial"]


cdef struct foi_buffer:
    int count
    int *index
    double *foi
    foi_buffer *next_buffer


cdef foi_buffer* allocate_foi_buffer(int buffer_size=4096) nogil:

    cdef int size = buffer_size
    cdef foi_buffer *buffer = <foi_buffer *> calloc(1, sizeof(foi_buffer))

    buffer[0].count = 0
    buffer[0].index = <int *> calloc(size, sizeof(int))
    buffer[0].foi = <double *> calloc(size, sizeof(double))
    buffer[0].next_buffer = <foi_buffer*>0

    return buffer


cdef foi_buffer* allocate_foi_buffers(int nthreads,
                                      int buffer_size=4096) nogil:
    cdef int size = buffer_size
    cdef int n = nthreads

    cdef foi_buffer *buffers = <foi_buffer *> calloc(n, sizeof(foi_buffer))

    for i in range(0, nthreads):
        buffers[i].count = 0
        buffers[i].index = <int *> calloc(size, sizeof(int))
        buffers[i].foi = <double *> calloc(size, sizeof(double))
        buffers[i].next_buffer = <foi_buffer*>0

    return buffers


cdef void free_foi_buffer(foi_buffer *buffer) nogil:
    if buffer:
        free(buffer[0].index)
        buffer[0].index = <int *>0
        free(buffer[0].foi)
        buffer[0].foi = <double *>0

        if buffer[0].next_buffer:
            free_foi_buffer(buffer[0].next_buffer)
            free(buffer[0].next_buffer)
            buffer[0].next_buffer = <foi_buffer *>0


cdef void free_foi_buffers(foi_buffer *buffers, int nthreads) nogil:
    cdef int n = nthreads

    for i in range(0, nthreads):
        free_foi_buffer(&(buffers[i]))

    free(buffers)


cdef void add_from_buffer(foi_buffer *buffer, double *wards_foi) nogil:
    cdef int i = 0
    cdef int count = buffer[0].count

    for i in range(0, count):
        wards_foi[buffer[0].index[i]] += buffer[0].foi[i]

    if buffer[0].next_buffer:
        add_from_buffer(buffer[0].next_buffer, wards_foi)

    # added all of buffer, so set count to 0 to empty
    buffer[0].count = 0


cdef inline void add_to_buffer(foi_buffer *buffer, int index, double value,
                               double *wards_foi,
                               int buffer_size=4096) nogil:

    cdef int count = buffer[0].count

    if buffer[0].count >= buffer_size:
        # we need to allocate another buffer and set to use that
        if not buffer[0].next_buffer:
            buffer[0].next_buffer = allocate_foi_buffer(buffer_size)

        add_to_buffer(buffer[0].next_buffer, index, value,
                      wards_foi, buffer_size)
    else:
        buffer[0].index[count] = index
        buffer[0].foi[count] = value
        buffer[0].count = count + 1


def advance_foi_omp(network: Network, population: Population,
                    infections: Infections, rngs,
                    nthreads: int, profiler: Profiler, **kwargs):
    """Advance the model calculating the new force of infection (foi)
       for all of the wards and links between wards, based on the
       current number of infections. Note that you must call this
       first before performing any other step in the iteration
       as this will update the foi based on the infections that
       occured the previous day. This is the parallel version of
       this function

       Parameters
       ----------
       network: Network
         The network being modelled
       population: Population
         The population experiencing the outbreak - contains the
         day number of the outbreak
       infections: Infections
         The space that holds all of the infections
       rngs:
         The list of thread-safe random number generators, one per thread
       nthreads: int
         The number of threads over which to parallelise the calculation
       profiler: Profiler
         The profiler used to profile this calculation
       kwargs:
         Extra arguments that may be used by other advancers, but which
         are not used by advance_play
    """

    links = network.links
    wards = network.nodes
    play = network.play
    params = network.params

    # Copy arguments from Python into C cdef variables
    cdef double uv = params.UV
    cdef double uvscale = population.scale_uv
    cdef int ts = population.day

    play_infections = infections.play
    infections = infections.work

    if uv > 0:
        uvscale *= (1.0-uv/2.0 + uv*cos(2.0*pi*ts/365.0)/2.0)

    cdef double * wards_day_foi = get_double_array_ptr(wards.day_foi)
    cdef double * wards_night_foi = get_double_array_ptr(wards.night_foi)

    cdef double * links_weight = get_double_array_ptr(links.weight)
    cdef double * play_weight = get_double_array_ptr(play.weight)

    cdef int * links_ifrom = get_int_array_ptr(links.ifrom)
    cdef int * links_ito = get_int_array_ptr(links.ito)

    cdef int * play_ifrom = get_int_array_ptr(play.ifrom)
    cdef int * play_ito = get_int_array_ptr(play.ito)

    cdef int * wards_begin_p = get_int_array_ptr(wards.begin_p)
    cdef int * wards_end_p = get_int_array_ptr(wards.end_p)

    cdef double * links_distance = get_double_array_ptr(links.distance)
    cdef double * play_distance = get_double_array_ptr(play.distance)

    cdef double cutoff = params.dyn_dist_cutoff

    # get the random number generator
    cdef uintptr_t [::1] rngs_view = rngs
    cdef binomial_rng* rng   # pointer to parallel rng

    # create and initialise variables used in the loop
    cdef int num_threads = nthreads
    cdef int thread_id = 0

    cdef int nnodes_plus_one = network.nnodes + 1
    cdef int nlinks_plus_one = network.nlinks + 1

    cdef int i = 0
    cdef int j = 0
    cdef int k = 0
    cdef int end_p = 0
    cdef int inf_ij = 0
    cdef int ifrom = 0
    cdef int ito = 0
    cdef int staying = 0
    cdef int moving = 0
    cdef int play_move = 0

    cdef int N_INF_CLASSES = len(infections)

    cdef double weight = 0.0
    cdef double cumulative_prob = 0.0
    cdef double prob_scaled = 0.0
    cdef double too_ill_to_move = 0.0
    cdef double scl_foi_uv = 0.0

    # allocate buffers that will be used to manage the reductions
    cdef foi_buffer * day_buffers = allocate_foi_buffers(num_threads)
    cdef foi_buffer * day_buffer
    cdef foi_buffer * night_buffers = allocate_foi_buffers(num_threads)
    cdef foi_buffer * night_buffer

    ## Finally(!) we can now declare the actual loop.
    ## This loops over all disease stages, and then in
    ## parallel over all wards and all links to then
    ## update the daytime and nighttime foi arrays
    ## for each ward (wards.day_foi and wards.night_foi)

    ## we begin by initialising the day and night fois to 0
    p = profiler.start("setup")
    for i in range(1, network.nnodes+1):
        wards_day_foi[i] = 0.0
        wards_night_foi[i] = 0.0
    p = p.stop()

    p = p.start("loop_over_classes")
    for i in range(0, N_INF_CLASSES):
        contrib_foi = params.disease_params.contrib_foi[i]
        beta = params.disease_params.beta[i]
        scl_foi_uv = contrib_foi * beta * uvscale
        too_ill_to_move = params.disease_params.too_ill_to_move[i]

        # number of people staying gets bigger as
        # PlayAtHome increases
        play_at_home_scl = <double>(params.dyn_play_at_home *
                                    too_ill_to_move)

        infections_i = get_int_array_ptr(infections[i])
        play_infections_i = get_int_array_ptr(play_infections[i])

        if contrib_foi > 0:
            p = p.start(f"work_{i}")
            with nogil, parallel(num_threads=num_threads):
                thread_id = cython.parallel.threadid()
                rng = _get_binomial_ptr(rngs_view[thread_id])
                day_buffer = &(day_buffers[thread_id])
                night_buffer = &(night_buffers[thread_id])
                day_buffer[0].count = 0
                night_buffer[0].count = 0

                for j in prange(1, nlinks_plus_one, schedule="static"):
                    # deterministic movements (e.g. to work)
                    inf_ij = infections_i[j]
                    if inf_ij > 0:
                        weight = links_weight[j]
                        ifrom = links_ifrom[j]
                        ito = links_ito[j]

                        if links_distance[j] < cutoff:
                            # number staying - this is G_ij
                            staying = _ran_binomial(rng,
                                                    too_ill_to_move,
                                                    inf_ij)

                            # number moving, this is I_ij - G_ij
                            moving = inf_ij - staying

                            if staying > 0:
                                add_to_buffer(day_buffer, ifrom,
                                              staying * scl_foi_uv,
                                              &(wards_day_foi[0]))

                            # Daytime Force of
                            # Infection is proportional to
                            # number of people staying
                            # in the ward (too ill to work)
                            # this is the sum for all G_ij (including g_ii
                            if moving > 0:
                                add_to_buffer(day_buffer, ito,
                                              moving * scl_foi_uv,
                                              &(wards_day_foi[0]))

                            # Daytime FOI for destination is incremented
                            # (including self links, I_ii)
                        else:
                            # outside cutoff
                            if inf_ij > 0:
                                add_to_buffer(day_buffer, ifrom,
                                              inf_ij * scl_foi_uv,
                                              &(wards_day_foi[0]))

                        if inf_ij > 0:
                            add_to_buffer(night_buffer, ifrom,
                                          inf_ij * scl_foi_uv,
                                          &(wards_night_foi[0]))
                        #wards_night_foi[ifrom] += inf_ij * scl_foi_uv

                        # Nighttime Force of Infection is
                        # prop. to the number of Infected individuals
                        # in the ward
                        # This I_ii in Lambda^N

                    # end of if inf_ij (are there any new infections)
                # end of infectious class loop
            # end of parallel section

            # do the reduction in serial in a deterministic order
            for j in range(0, num_threads):
                add_from_buffer(&(day_buffers[j]), &(wards_day_foi[0]))
                add_from_buffer(&(night_buffers[j]), &(wards_night_foi[0]))

            p = p.stop()

            p = p.start(f"play_{i}")
            with nogil, parallel(num_threads=num_threads):
                thread_id = cython.parallel.threadid()
                rng = _get_binomial_ptr(rngs_view[thread_id])
                day_buffer = &(day_buffers[thread_id])
                day_buffer[0].count = 0

                for j in prange(1, nnodes_plus_one, schedule="static"):
                    # playmatrix loop FOI loop (random/unpredictable movements)
                    inf_ij = play_infections_i[j]
                    if inf_ij > 0:
                        wards_night_foi[j] += inf_ij * scl_foi_uv

                        staying = _ran_binomial(rng, play_at_home_scl, inf_ij)
                        moving = inf_ij - staying

                        cumulative_prob = 0.0
                        k = wards_begin_p[j]

                        end_p = wards_end_p[j]

                        while (moving > 0) and (k < end_p):
                            # distributing people across play wards
                            if play_distance[k] < cutoff:
                                weight = play_weight[k]
                                ifrom = play_ifrom[k]
                                ito = play_ito[k]

                                prob_scaled = weight / (1.0 - cumulative_prob)
                                cumulative_prob = cumulative_prob + weight

                                play_move = _ran_binomial(rng, prob_scaled,
                                                          moving)

                                add_to_buffer(day_buffer, ito,
                                              play_move * scl_foi_uv,
                                              &(wards_day_foi[0]))

                                moving = moving - play_move
                            # end of if within cutoff

                            k = k + 1
                        # end of while loop

                        wards_day_foi[j] += (moving + staying) * scl_foi_uv
                    # end of if inf_ij (there are new infections)

                # end of loop over all nodes
            # end of parallel

            # perform the reduction in series in a predictable order
            for j in range(0, num_threads):
                add_from_buffer(&(day_buffers[j]), &(wards_day_foi[0]))

            p = p.stop()
        # end of params.disease_params.contrib_foi[i] > 0:
    p = p.stop()
    # end of loop over all disease classes

    free_foi_buffers(&(day_buffers[0]), num_threads)
    free_foi_buffers(&(night_buffers[0]), num_threads)


def advance_foi_serial(network: Network, population: Population,
                       infections: Infections, rngs,
                       profiler: Profiler, **kwargs):
    """Advance the model calculating the new force of infection (foi)
       for all of the wards and links between wards, based on the
       current number of infections. Note that you must call this
       first before performing any other step in the iteration
       as this will update the foi based on the infections that
       occured the previous day. This is the serial version of
       this function

       Parameters
       ----------
       network: Network
         The network being modelled
       population: Population
         The population experiencing the outbreak - contains the
         day number of the outbreak
       infections: Infections
         The space that holds all of the "work" infections
       rngs:
         The list of thread-safe random number generators, one per thread
       day: int
         The day of the outbreak (timestep in the simulation)
       profiler: Profiler
         The profiler used to profile this calculation
       kwargs:
         Extra arguments that may be used by other advancers, but which
         are not used by advance_play
    """

    links = network.links
    wards = network.nodes
    play = network.play
    params = network.params

    # Copy arguments from Python into C cdef variables
    cdef double uv = params.UV
    cdef double uvscale = population.scale_uv
    cdef int ts = population.day

    play_infections = infections.play
    infections = infections.work

    if uv > 0:
        uvscale *= (1.0-uv/2.0 + uv*cos(2.0*pi*ts/365.0)/2.0)

    cdef double * wards_day_foi = get_double_array_ptr(wards.day_foi)
    cdef double * wards_night_foi = get_double_array_ptr(wards.night_foi)

    cdef double * links_weight = get_double_array_ptr(links.weight)
    cdef double * play_weight = get_double_array_ptr(play.weight)

    cdef int * links_ifrom = get_int_array_ptr(links.ifrom)
    cdef int * links_ito = get_int_array_ptr(links.ito)

    cdef int * play_ifrom = get_int_array_ptr(play.ifrom)
    cdef int * play_ito = get_int_array_ptr(play.ito)

    cdef int * wards_begin_p = get_int_array_ptr(wards.begin_p)
    cdef int * wards_end_p = get_int_array_ptr(wards.end_p)

    cdef double * links_distance = get_double_array_ptr(links.distance)
    cdef double * play_distance = get_double_array_ptr(play.distance)

    cdef double cutoff = params.dyn_dist_cutoff

    # get the random number generator
    cdef binomial_rng* rng = _get_binomial_ptr(rngs[0])

    cdef int nnodes_plus_one = network.nnodes + 1
    cdef int nlinks_plus_one = network.nlinks + 1

    cdef int i = 0
    cdef int j = 0
    cdef int k = 0
    cdef int end_p = 0
    cdef int inf_ij = 0
    cdef int ifrom = 0
    cdef int ito = 0
    cdef int staying = 0
    cdef int moving = 0
    cdef int play_move = 0

    cdef int N_INF_CLASSES = len(infections)

    cdef double weight = 0.0
    cdef double cumulative_prob = 0.0
    cdef double prob_scaled = 0.0
    cdef double too_ill_to_move = 0.0
    cdef double scl_foi_uv = 0.0

    ## Finally(!) we can now declare the actual loop.
    ## This loops over all disease stages, and then in
    ## parallel over all wards and all links to then
    ## update the daytime and nighttime foi arrays
    ## for each ward (wards.day_foi and wards.night_foi)

    ## we begin by initialising the day and night fois to 0
    p = profiler.start("setup")
    for i in range(1, network.nnodes+1):
        wards_day_foi[i] = 0.0
        wards_night_foi[i] = 0.0
    p = p.stop()

    p = p.start("loop_over_classes")
    for i in range(0, N_INF_CLASSES):
        contrib_foi = params.disease_params.contrib_foi[i]
        beta = params.disease_params.beta[i]
        scl_foi_uv = contrib_foi * beta * uvscale
        too_ill_to_move = params.disease_params.too_ill_to_move[i]

        # number of people staying gets bigger as
        # PlayAtHome increases
        play_at_home_scl = <double>(params.dyn_play_at_home *
                                    too_ill_to_move)

        infections_i = get_int_array_ptr(infections[i])
        play_infections_i = get_int_array_ptr(play_infections[i])

        if contrib_foi > 0:
            p = p.start(f"work_{i}")
            with nogil:
                for j in range(1, nlinks_plus_one):
                    # deterministic movements (e.g. to work)
                    inf_ij = infections_i[j]
                    if inf_ij > 0:
                        weight = links_weight[j]
                        ifrom = links_ifrom[j]
                        ito = links_ito[j]

                        if links_distance[j] < cutoff:
                            # number staying - this is G_ij
                            staying = _ran_binomial(rng,
                                                    too_ill_to_move,
                                                    inf_ij)

                            # number moving, this is I_ij - G_ij
                            moving = inf_ij - staying

                            if staying > 0:
                                wards_day_foi[ifrom] += staying * scl_foi_uv

                            # Daytime Force of
                            # Infection is proportional to
                            # number of people staying
                            # in the ward (too ill to work)
                            # this is the sum for all G_ij (including g_ii
                            if moving > 0:
                                wards_day_foi[ito] += moving * scl_foi_uv

                            # Daytime FOI for destination is incremented
                            # (including self links, I_ii)
                        else:
                            # outside cutoff
                            if inf_ij > 0:
                                wards_day_foi[ifrom] += inf_ij * scl_foi_uv

                        if inf_ij > 0:
                            wards_night_foi[ifrom] += inf_ij * scl_foi_uv

                        # Nighttime Force of Infection is
                        # prop. to the number of Infected individuals
                        # in the ward
                        # This I_ii in Lambda^N

                    # end of if inf_ij (are there any new infections)
                # end of infectious class loop
            # end of nogil
            p = p.stop()

            p = p.start(f"play_{i}")
            with nogil:
                for j in range(1, nnodes_plus_one):
                    # playmatrix loop FOI loop (random/unpredictable movements)
                    inf_ij = play_infections_i[j]
                    if inf_ij > 0:
                        wards_night_foi[j] += inf_ij * scl_foi_uv

                        staying = _ran_binomial(rng, play_at_home_scl, inf_ij)
                        moving = inf_ij - staying

                        cumulative_prob = 0.0
                        k = wards_begin_p[j]

                        end_p = wards_end_p[j]

                        while (moving > 0) and (k < end_p):
                            # distributing people across play wards
                            if play_distance[k] < cutoff:
                                weight = play_weight[k]
                                ifrom = play_ifrom[k]
                                ito = play_ito[k]

                                prob_scaled = weight / (1.0 - cumulative_prob)
                                cumulative_prob = cumulative_prob + weight

                                play_move = _ran_binomial(rng, prob_scaled,
                                                          moving)

                                wards_day_foi[ito] += play_move * scl_foi_uv

                                moving = moving - play_move
                            # end of if within cutoff

                            k = k + 1
                        # end of while loop

                        wards_day_foi[j] += (moving + staying) * scl_foi_uv
                    # end of if inf_ij (there are new infections)

                # end of loop over all nodes
            # end of nogil
            p = p.stop()
        # end of params.disease_params.contrib_foi[i] > 0:
    p = p.stop()
    # end of loop over all disease classes


def advance_foi(nthreads: int, **kwargs):
    """Advance the model calculating the new force of infection (foi)
       for all of the wards and links between wards, based on the
       current number of infections. Note that you must call this
       first before performing any other step in the iteration
       as this will update the foi based on the infections that
       occured the previous day. This is the parallel version of
       this function

       Parameters
       ----------
       network: Network or Networks
         The network being modelled
       population: Population
         The population experiencing the outbreak - contains the
         day number of the outbreak
       infections: Infections
         The space that holds all of the infections
       rngs:
         The list of thread-safe random number generators, one per thread
       nthreads: int
         The number of threads over which to parallelise the calculation
       profiler: Profiler
         The profiler used to profile this calculation
       kwargs:
         Extra arguments that may be used by other advancers, but which
         are not used by advance_play
    """
    call_function_on_network(nthreads=nthreads,
                             func=advance_foi_serial,
                             parallel=advance_foi_omp,
                             switch_to_parallel=5,
                             **kwargs)



"""
.. currentmodule:: metawards.analysis

Functions
=========

.. autosummary::
    :toctree: generated/

    animate_plots
    create_average_plot
    create_overview_plot
    import_animate_modules
    import_graphics_modules
    save_summary_plots

"""

from ._animate_plots import *
from ._summary_plot import *




from dataclasses import dataclass as _dataclass
from dataclasses import field as _field
from typing import List as _List

from ._network import Network
from ._parameters import Parameters
from ._demographics import Demographics
from ._population import Population
from ._outputfiles import OutputFiles

__all__ = ["Networks"]


@_dataclass
class Networks:
    """This is a combination of Network objects which together represent
       an entire diverse population. Each individual Network is used to
       model the disease outbreak within a single demographic of the
       population. Multiple demographics are modelled by combining
       multiple networks. Special merge functions enable joint
       FOIs to be calculated, through which an outbreak in one
       network can cross-infect a demographic in another network.

       The Networks can be very independent, and don't necessarily
       need to have the same links. However, it is assumed (and checked)
       that each network will have the same nodes.
    """

    #: The overall Network, which contains a combination of all of the
    #: sub-networks. This is used for summary analysis and also as
    #: a means of merging and distributing data between sub-networks
    overall: Network = None

    #: The list of Networks, one for each demographic, ordered in the
    #: same order as the "Demographics" object. This is empty if
    #: only a single demographic is modelled
    subnets: _List[Network] = _field(default_factory=list)

    #: Metadata about each of the demographics being modelled. This is
    #: None if only a single demographic is modelled
    demographics: Demographics = None

    @property
    def params(self) -> int:
        """The overall parameters that are then specialised for the
           different demographics
        """
        if self.overall is not None:
            return self.overall.params
        else:
            return None

    def assert_sane(self, profiler: None):
        """Assert that the networks is sane. This checks that the
           networks and all of the demographic sub-networks are
           laid out correctly in memory and that they don't have
           anything unexpected. Checking here will prevent us from having
           to check every time the networks are accessed
        """
        if self.overall:
            self.overall.assert_sane(profiler=profiler)

        for subnet in self.subnets:
            subnet.assert_sane(profiler=profiler)

        # SHOULD ASSERT HERE THAT THE POPULATIONS OF ALL OF THE SUBNETS
        # IN EACH WARD SUM UP TO THE POPULATION IN THE OVERALL NETWORK
        # WARDS

    @staticmethod
    def build(network: Network, demographics: Demographics,
              profiler=None, nthreads: int = 1):
        """Build the set of networks that will model the passed
           demographics based on the overall population model
           in the passed network

           Parameters
           ----------
           network: Network
             The overall population model - this contains the base
             parameters, wards, work and play links that define
             the model outbreak
           demographics: Demographics
             Information about each of the demographics to be modelled.
             Note that the sum of the "work" and "play" populations
             across all demographics must be 1.0 in all wards in
             the model
           profiler: Profiler
             Optional profiler used to profile this build
           nthreads: int
             Number of threads over which to distribute the work

           Returns
           -------
           networks: Networks
             The set of Networks that represent the model run over the
             full set of different demographics
        """
        if not isinstance(network, Network):
            raise TypeError(f"You can only specialise a Network")

        if demographics is None or len(demographics) < 2:
            raise ValueError(f"You can only create a Networks object "
                             f"with a valid Demographics that contains "
                             f"more than one demographic")

        if profiler is None:
            from .utils._profiler import NullProfiler
            profiler = NullProfiler()

        p = profiler.start("specialise")

        subnets = []

        # specialise the network for each demographic
        for i in range(0, len(demographics)):
            p = p.start(f"demographic_{i}")
            subnets.append(network.specialise(demographic=demographics[i],
                                              profiler=p,
                                              nthreads=nthreads))
            p = p.stop()

        p = p.start("distribute_remainders")
        from .utils._scale_susceptibles import distribute_remainders
        distribute_remainders(network=network, subnets=subnets,
                              profiler=p, nthreads=nthreads,
                              random_seed=demographics.random_seed)
        p = p.stop()

        total_pop = network.population
        sum_pop = 0

        from .utils._console import Console

        Console.print(f"Specialising network - population: {total_pop}")

        for i, subnet in enumerate(subnets):
            pop = subnet.population
            sum_pop += pop
            Console.print(f"  {demographics[i].name} - population: {pop}")

        if total_pop != sum_pop:
            raise AssertionError(
                f"The sum of the population of the demographic "
                f"sub-networks ({sum_pop}) does not equal the population "
                f"of the total network ({total_pop}). This is a bug!")

        result = Networks()
        result.overall = network
        result.subnets = subnets
        result.demographics = demographics

        p = p.stop()

        return result

    def copy(self):
        """Return a copy of this Networks. Use this to hold a copy of
           the networks that you can use to reset between runs
        """
        from copy import copy
        networks = copy(self)
        networks.overall = self.overall.copy()

        subnets = []

        for subnet in self.subnets:
            subnets.append(subnet.copy())

        networks.subnets = subnets

        networks.demographics = self.demographics.copy()

        return networks

    def aggregate(self, profiler=None, nthreads: int = 1):
        """Aggregate all of the sub-network population infection data
           so that this is available in the overall network
        """
        from .utils._aggregate import aggregate_networks
        aggregate_networks(network=self, profiler=profiler, nthreads=nthreads)

    def run(self, population: Population,
            output_dir: OutputFiles,
            seed: int = None,
            nsteps: int = None,
            nthreads: int = None,
            iterator=None,
            extractor=None,
            mover=None,
            mixer=None,
            profiler=None) -> Population:
        """Run the model simulation for the passed population.
           The random number seed is given in 'seed'. If this
           is None, then a random seed is used.

           All output files are written to 'output_dir'

           The simulation will continue until the infection has
           died out or until 'nsteps' has passed (keep as 'None'
           to prevent exiting early).

           Parameters
           ----------
           population: Population
             The initial population at the start of the model outbreak.
             This is also used to set start date and day of the model
             outbreak
           output_dir: OutputFiles
             The directory to write all of the output into
           seed: int
             The random number seed used for this model run. If this is
             None then a very random random number seed will be used
           nsteps: int
             The maximum number of steps to run in the outbreak. If None
             then run until the outbreak has finished
           profiler: Profiler
             The profiler to use - a new one is created if one isn't passed
           nthreads: int
             Number of threads over which to parallelise this model run
           iterator: function
             Function that is called at each iteration to get the functions
             that are used to advance the model
           extractor: function
             Function that is called at each iteration to get the functions
             that are used to extract data for analysis or writing to files
           mixer: function
             Function that is called to mix the data calculated for each
             of the sub-networks for the different demographics and
             merge it together so that this is shared
           mover: function
             Function that is called to move the population between
             different demographics

           Returns
           -------
           population: Population
             The final population at the end of the run
        """
        # Create the random number generator
        from .utils._ran_binomial import seed_ran_binomial, ran_binomial

        if seed == 0:
            # this is a special mode that a developer can use to force
            # all jobs to use the same random number seed (15324) that
            # is used for comparing outputs. This should NEVER be used
            # for production code
            from .utils._console import Console
            Console.warning("Using special mode to fix all random number "
                            "seeds to 15324. DO NOT USE IN PRODUCTION!!!")
            rng = seed_ran_binomial(seed=15324)
        else:
            rng = seed_ran_binomial(seed=seed)

        # Print the first five random numbers so that we can
        # compare to other codes/runs, and be sure that we are
        # generating the same random sequence
        randnums = []
        for i in range(0, 5):
            randnums.append(str(ran_binomial(rng, 0.5, 100)))

        from .utils._console import Console
        Console.print(f"First five random numbers equal {', '.join(randnums)}")
        randnums = None

        if nthreads is None:
            from .utils._parallel import get_available_num_threads
            nthreads = get_available_num_threads()

        Console.print(f"Number of threads used equals {nthreads}")

        from .utils._parallel import create_thread_generators
        rngs = create_thread_generators(rng, nthreads)

        # Create space to hold the results of the simulation
        Console.print("Initialise infections...")
        infections = self.initialise_infections()

        from .utils import run_model
        population = run_model(network=self,
                               population=population,
                               infections=infections,
                               rngs=rngs, output_dir=output_dir,
                               nsteps=nsteps,
                               nthreads=nthreads,
                               profiler=profiler,
                               iterator=iterator, extractor=extractor,
                               mixer=mixer, mover=mover)

        return population

    def reset_everything(self, nthreads: int = 1, profiler=None):
        """Resets the networks ready for a new run of the model"""
        if self.overall:
            self.overall.reset_everything(nthreads=nthreads, profiler=profiler)

        for subnet in self.subnets:
            subnet.reset_everything(nthreads=nthreads, profiler=profiler)

    def update(self, params: Parameters, demographics=None,
               nthreads: int = 1, profiler=None):
        """Update this network with a new set of parameters
           (and optionally demographics).

           This is used to update the parameters for the network
           for a new run. The network will be reset
           and ready for a new run.

           Parameters
           ----------
           params: Parameters
             The new parameters with which to update this Network
           demographics: Demographics
             The new demographics with which to update this Network.
             Note that this will return a Network object that contains
             the specilisation of this Network
           nthreads: int
             Number of threads over which to parallelise this update
           profiler: Profiler
             The profiler used to profile this update

           Returns
           -------
           network: Network or Networks
             Either this Network after it has been updated, or the
             resulting Networks from specialising this Network using
             Demographics
        """
        if profiler is None:
            from .utils import NullProfiler
            profiler = NullProfiler()

        p = profiler.start("overall.update")
        self.overall.update(params, profiler=p)
        p = p.stop()

        if demographics is not None:
            if demographics != self.demographics:
                # we have a change in demographics, so need to re-specialise
                networks = demographics.specialise(network=self.overall,
                                                   profiler=p,
                                                   nthreads=nthreads)
                p.stop()
                return networks

        for i in range(0, len(self.demographics)):
            demographic = self.demographics[i]

            # TODO::: RESET THE SUSCEPTIBLE POPULATIONS BACK TO
            #         WHERE THEY SHOULD BE AT THE START OF THE RUN

            p = p.start(f"{demographic.name}.update")
            if demographic.adjustment:
                subnet_params = params.set_variables(demographic.adjustment)
            else:
                subnet_params = params

            self.subnets[i].update(subnet_params, profiler=p)
            p = p.stop()

    def initialise_infections(self, nthreads: int = 1):
        """Initialise and return the space that will be used
           to track infections
        """
        from ._infections import Infections
        return Infections.build(network=self)

    def rescale_play_matrix(self, nthreads: int = 1, profiler=None):
        """Rescale the play matrix"""
        if self.overall:
            self.overall.reset_everything(nthreads=nthreads, profiler=profiler)

        for subnet in self.subnets:
            subnet.rescale_play_matrix(nthreads=nthreads, profiler=profiler)

    def move_from_play_to_work(self, nthreads: int = 1, profiler=None):
        """Move the population from play to work"""
        if self.overall:
            self.overall.move_from_play_to_work(nthreads=nthreads,
                                                profiler=profiler)

        for subnet in self.subnets:
            subnet.move_from_play_to_work(nthreads=nthreads,
                                          profiler=profiler)




import time

__all__ = ["Profiler", "NullProfiler"]


class NullProfiler:
    """This is a null profiler that does nothing"""

    def __init__(self, name: str = None, parent=None):
        pass

    def is_null(self) -> bool:
        return True

    def __str__(self):
        return "No profiling data collected"

    def start(self, name: str = None):
        return self

    def stop(self):
        return self


class Profiler:
    """This is a simple profiling class that supports manual
       instrumenting of the code. It is used for sub-function
       profiling.
    """

    def __init__(self, name: str = None, parent=None):
        self._name = name
        self._parent = parent
        self._children = []
        self._start = None
        self._end = None

    def is_null(self) -> bool:
        """Return whether this is a null profiler"""
        return False

    def _to_string(self):
        """Used to write the results of profiling as a report"""
        lines = []

        t = self.total()

        if self._name is None:
            if self._parent is None:
                self._name = "Total time"

        if t:
            if len(self._children) > 0:
                ct = self.child_total()

                if ct:
                    lines.append("%s: %.3f ms (%.3f ms)" % (self._name, t, ct))
                else:
                    lines.append("%s: %.3f ms (???) ms" % (self._name, t))
            else:
                lines.append("%s: %.3f ms" % (self._name, t))

        elif self._start is None:
            lines.append(f"{self._name}")
        else:
            lines.append(f"{self._name}: still timing...")

        for child in self._children:
            clines = child._to_string()
            lines.append(f"  \\-{clines[0]}")

            if len(clines) > 1:
                for l in clines[1:]:
                    lines.append(f"    {l}")

        return lines

    def __str__(self):
        """Return the results of profiling as a report"""
        return "\n".join(self._to_string())

    def name(self) -> str:
        """Return the name of this profiler"""
        return self._name

    def child_total(self) -> float:
        """Return the sum of time spent in the child profilers"""
        sum = None

        for child in self._children:
            t = child.total()

            if t is not None:
                if sum is None:
                    sum = t
                else:
                    sum += t

        return sum

    def total(self) -> float:
        """Return the amount of time that was recorded for this
           profiler in milliseconds (accurate to ~nanoseconds)
        """
        if self._parent is None:
            # this is the top-level profiler - just return the child total
            return self.child_total()
        elif self._end:
            return (self._end - self._start) * 0.000001    # ns to ms
        else:
            return None

    def start(self, name: str):
        """Start profiling the section called 'name'. This
           returns the updated profiler, e.g.

           p = Profiler()

           p = p.start("long_loop")

           # run some code

           p = p.end()

           print(p)
        """
        p = Profiler(name=name, parent=self)
        self._children.append(p)

        p._start = time.time_ns()
        return p

    def stop(self):
        """Stop profiling. This records the end time
           and returns the parent profiler (if we have one)
        """
        end = time.time_ns()

        if self._start is None:
            from ._console import Console
            Console.warning(f"You cannot stop profiler {self._name} as "
                            f"it has not been started!")
        elif self._end is not None:
            from ._console import Console
            Console.warning(
                f"WARNING: You cannot stop profiler {self._name} as "
                f"it has already been stopped!")
        else:
            self._end = end

        if self._parent:
            return self._parent
        else:
            # no parent - just return this profiler
            return self




from typing import List as _List
from ..utils._get_functions import MetaFunction

__all__ = ["extract_default"]


def extract_default(stage: str, **kwargs) -> _List[MetaFunction]:
    """This returns the default list of 'output_XXX' functions that
       are called in sequence for each iteration of the model run.
       These functions are used to output data to files for
       future processing

       Parameters
       ----------
       stage: str
         Which stage of the day is to be modelled

       Returns
       -------
       funcs: List[MetaFunction]
         The list of functions that will be called in sequence
    """

    if stage == "initialise":
        from ._output_core import setup_core, output_core
        return [setup_core, output_core]

    elif stage == "infect":
        # this must be called at the end of the 'infect' stage to
        # accumulate all of the new infections
        from ._output_core import output_core
        return [output_core]

    elif stage == "analyse":
        from ._output_basic import output_basic
        from ._output_dispersal import output_dispersal
        from ._output_incidence import output_incidence
        from ._output_prevalence import output_prevalence

        return [output_basic, output_dispersal,
                output_incidence, output_prevalence]

    elif stage == "finalise":
        from ._output_trajectory import output_trajectory
        return [output_trajectory]

    elif stage == "summary":
        # output the summary results.csv.bz2 file
        from ._output_final_report import output_final_report
        return [output_final_report]

    else:
        # we don't do anything at the "foi", "analyse" or "finalise" stages
        return []




from .._network import Network
from .._population import Population
from .._outputfiles import OutputFiles
from .._workspace import Workspace

from ..utils._get_functions import call_function_on_network

__all__ = ["output_basic"]


def output_basic_serial(network: Network, population: Population,
                        output_dir: OutputFiles,
                        workspace: Workspace,
                        **kwargs):
    """This will write basic trajectory data to the output
       files. This will be the number of infected wards,
       total infections, play infections and work infections
       for each disease stage for each timestep

       Parameters
       ----------
       network: Network
         The network over which the outbreak is being modelled
       population: Population
         The population experiencing the outbreak
       output_dir: OutputFiles
         The directory in which to place all output files
       workspace: Workspace
         A workspace that can be used to extract data
       kwargs
         Extra argumentst that are ignored by this function
    """

    if network.name is None:
        name = ""
    else:
        name = "_" + network.name.replace(" ", "_")

    # get the file handles - this will open the files if
    # they have not already been created
    n_inf_wards_file = output_dir.open(f"NumberWardsInfected{name}.dat")
    total_file = output_dir.open(f"TotalInfections{name}.dat")
    work_file = output_dir.open(f"WorkInfections{name}.dat")
    play_file = output_dir.open(f"PlayInfections{name}.dat")

    ts = f"{population.day} "

    def _join(array):
        return " ".join([str(x) for x in array])

    total_file.write(str(population.total) + "\n")

    n_inf_wards_file.write(ts + _join(workspace.n_inf_wards) + "\n")
    work_file.write(ts + _join(workspace.inf_tot) + "\n")
    play_file.write(ts + _join(workspace.pinf_tot) + "\n")


def output_basic(nthreads: int = 1, **kwargs):
    """This will write basic trajectory data to the output
       files. This will be the number of infected wards,
       total infections, play infections and work infections
       for each disease stage for each timestep

       Parameters
       ----------
       network: Network
         The network over which the outbreak is being modelled
       population: Population
         The population experiencing the outbreak
       output_dir: OutputFiles
         The directory in which to place all output files
       workspace: Workspace
         A workspace that can be used to extract data
       kwargs
         Extra argumentst that are ignored by this function
    """
    call_function_on_network(nthreads=1,
                             func=output_basic_serial,
                             call_on_overall=True,
                             **kwargs)



#!/bin/env/python3
#cython: linetrace=False
# MUST ALWAYS DISABLE AS WAY TOO SLOW FOR ITERATE

cimport cython
from cython.parallel import parallel, prange
from libc.stdint cimport uintptr_t

from .._network import Network

from ..utils._profiler import Profiler
from ..utils._get_functions import call_function_on_network

from ..utils._ran_binomial cimport _ran_binomial, \
                                   _get_binomial_ptr, binomial_rng

from ..utils._get_array_ptr cimport get_int_array_ptr, get_double_array_ptr

__all__ = ["advance_fixed", "advance_fixed_omp",
           "advance_fixed_serial"]


def advance_fixed_omp(network: Network, infections, rngs,
                      nthreads: int, profiler: Profiler, **kwargs):
    """Advance the model by triggering infections related to fixed
       'work' movements (parallel version of the function)

       Parameters
       ----------
       network: Network
         The network being modelled
       infections:
         The space that holds all of the "work" infections
       rngs:
         The list of thread-safe random number generators, one per thread
       nthreads: int
         The number of threads over which to parallelise the calculation
       profiler: Profiler
         The profiler used to profile this calculation
       kwargs:
         Extra arguments that may be used by other advancers, but which
         are not used by advance_play
    """

    links = network.links
    wards = network.nodes
    params = network.params

    infections = infections.work

    # Copy arguments from Python into C cdef variables
    cdef int * links_ifrom = get_int_array_ptr(links.ifrom)
    cdef int * links_ito = get_int_array_ptr(links.ito)

    cdef double * links_distance = get_double_array_ptr(links.distance)
    cdef double * links_suscept = get_double_array_ptr(links.suscept)

    cdef double * wards_day_foi = get_double_array_ptr(wards.day_foi)
    cdef double * wards_day_inf_prob = get_double_array_ptr(wards.day_inf_prob)
    cdef double * wards_night_inf_prob = get_double_array_ptr(
                                                        wards.night_inf_prob)

    cdef double cutoff = params.dyn_dist_cutoff

    # Pointer to the infections array - only need [0] as this loop
    # is creating new infections
    cdef int * infections_i = get_int_array_ptr(infections[0])

    # get the random number generator
    cdef uintptr_t [::1] rngs_view = rngs
    cdef binomial_rng* rng   # pointer to parallel rng

    # create and initialise variables used in the loop
    cdef int num_threads = nthreads
    cdef int thread_id = 0
    cdef int j = 0
    cdef int l = 0
    cdef int ifrom = 0
    cdef int ito = 0
    cdef int nlinks_plus_one = network.nlinks + 1

    cdef double inf_prob = 0.0
    cdef double distance = 0.0

    ## Finally(!) we can now declare the actual loop.
    ## This loops in parallel over all links between
    ## wards to create new infections that appear in
    ## those wards at daytime and nighttime

    p = profiler.start("fixed")
    with nogil, parallel(num_threads=num_threads):
        thread_id = cython.parallel.threadid()
        rng = _get_binomial_ptr(rngs_view[thread_id])

        for j in prange(1, nlinks_plus_one, schedule="static"):
            # actual new infections for fixed movements
            inf_prob = 0

            ifrom = links_ifrom[j]
            ito = links_ito[j]
            distance = links_distance[j]

            if distance < cutoff:
                # distance is below cutoff (reasonable distance)
                # infect in work ward
                if wards_day_foi[ito] > 0:
                    # daytime infection of link j
                    inf_prob = wards_day_inf_prob[ito]

                # end of if wards.day_foi[ito] > 0
            # end of if distance < cutoff
            elif wards_day_foi[ifrom] > 0:
                # if distance is too large then infect in home ward with day FOI
                inf_prob = wards_day_inf_prob[ifrom]

            if inf_prob > 0.0:
                # daytime infection of workers
                l = _ran_binomial(rng, inf_prob, <int>(links_suscept[j]))

                if l > 0:
                    # actual infection
                    #print(f"InfProb {inf_prob}, susc {links.suscept[j]}, l {l}")
                    infections_i[j] += l
                    links_suscept[j] -= l

            # nighttime infection of workers
            inf_prob = wards_night_inf_prob[ifrom]

            if inf_prob > 0.0:
                l = _ran_binomial(rng, inf_prob, <int>(links_suscept[j]))

                #if l > links_suscept[j]:
                #    print(f"l > links[{j}].suscept {links_suscept[j]} nighttime")

                if l > 0:
                    # actual infection
                    # print(f"NIGHT InfProb {inf_prob}, susc {links.suscept[j]}, {l}")
                    infections_i[j] += l
                    links_suscept[j] -= l
            # end of wards.night_foi[ifrom] > 0  (nighttime infections)
        # end of loop over all network links
    # end of parallel section
    p = p.stop()


def advance_fixed_serial(network: Network, infections, rngs,
                         profiler: Profiler, **kwargs):
    """Advance the model by triggering infections related to fixed
       'work' movements (serial version of the function)

       Parameters
       ----------
       network: Network
         The network being modelled
       infections:
         The space that holds all of the "work" infections
       rngs:
         The list of thread-safe random number generators, one per thread
       profiler: Profiler
         The profiler used to profile this calculation
       kwargs:
         Extra arguments that may be used by other advancers, but which
         are not used by advance_play
    """

    links = network.links
    wards = network.nodes
    params = network.params

    infections = infections.work

    # Copy arguments from Python into C cdef variables
    cdef int * links_ifrom = get_int_array_ptr(links.ifrom)
    cdef int * links_ito = get_int_array_ptr(links.ito)

    cdef double * links_distance = get_double_array_ptr(links.distance)
    cdef double * links_suscept = get_double_array_ptr(links.suscept)

    cdef double * wards_day_foi = get_double_array_ptr(wards.day_foi)
    cdef double * wards_day_inf_prob = get_double_array_ptr(wards.day_inf_prob)
    cdef double * wards_night_inf_prob = get_double_array_ptr(
                                                        wards.night_inf_prob)

    cdef double cutoff = params.dyn_dist_cutoff

    # Pointer to the infections array - only need [0] as this loop
    # is creating new infections
    cdef int * infections_i = get_int_array_ptr(infections[0])

    # get the random number generator
    cdef binomial_rng* rng = _get_binomial_ptr(rngs[0])

    # create and initialise variables used in the loop
    cdef int j = 0
    cdef int l = 0
    cdef int ifrom = 0
    cdef int ito = 0
    cdef int nlinks_plus_one = network.nlinks + 1

    cdef double inf_prob = 0.0
    cdef double distance = 0.0

    ## Finally(!) we can now declare the actual loop.
    ## This loops in parallel over all links between
    ## wards to create new infections that appear in
    ## those wards at daytime and nighttime

    p = profiler.start("fixed")
    with nogil:
        for j in range(1, nlinks_plus_one):
            # actual new infections for fixed movements
            inf_prob = 0

            ifrom = links_ifrom[j]
            ito = links_ito[j]
            distance = links_distance[j]

            if distance < cutoff:
                # distance is below cutoff (reasonable distance)
                # infect in work ward
                if wards_day_foi[ito] > 0:
                    # daytime infection of link j
                    inf_prob = wards_day_inf_prob[ito]

                # end of if wards.day_foi[ito] > 0
            # end of if distance < cutoff
            elif wards_day_foi[ifrom] > 0:
                # if distance is too large then infect in home ward with day FOI
                inf_prob = wards_day_inf_prob[ifrom]

            if inf_prob > 0.0:
                # daytime infection of workers
                l = _ran_binomial(rng, inf_prob, <int>(links_suscept[j]))

                if l > 0:
                    # actual infection
                    #print(f"InfProb {inf_prob}, susc {links.suscept[j]}, l {l}")
                    infections_i[j] += l
                    links_suscept[j] -= l

            # nighttime infection of workers
            inf_prob = wards_night_inf_prob[ifrom]

            if inf_prob > 0.0:
                l = _ran_binomial(rng, inf_prob, <int>(links_suscept[j]))

                #if l > links_suscept[j]:
                #    print(f"l > links[{j}].suscept {links_suscept[j]} nighttime")

                if l > 0:
                    # actual infection
                    # print(f"NIGHT InfProb {inf_prob}, susc {links.suscept[j]}, {l}")
                    infections_i[j] += l
                    links_suscept[j] -= l
            # end of wards.night_foi[ifrom] > 0  (nighttime infections)
        # end of loop over all network links
    # end of parallel section
    p = p.stop()


def advance_fixed(nthreads: int, **kwargs):
    """Advance the model by triggering infections related to fixed
       'work' movements (parallel version of the function)

       Parameters
       ----------
       network: Network
         The network being modelled
       infections:
         The space that holds all of the "work" infections
       rngs:
         The list of thread-safe random number generators, one per thread
       nthreads: int
         The number of threads over which to parallelise the calculation
       profiler: Profiler
         The profiler used to profile this calculation
       kwargs:
         Extra arguments that may be used by other advancers, but which
         are not used by advance_play
    """
    call_function_on_network(nthreads=nthreads,
                             func=advance_fixed_serial,
                             parallel=advance_fixed_omp,
                             **kwargs)




from typing import Union as _Union

from .._network import Network
from .._networks import Networks
from .._infections import Infections
from .._outputfiles import OutputFiles
from .._workspace import Workspace
from .._population import Population, Populations
from ._profiler import Profiler
from ._get_functions import get_initialise_functions, \
    get_model_loop_functions, \
    get_finalise_functions, \
    MetaFunction, \
    accepts_stage

__all__ = ["run_model"]


def run_model(network: _Union[Network, Networks],
              infections: Infections,
              rngs,
              output_dir: OutputFiles,
              population: Population = Population(initial=57104043),
              nsteps: int = None,
              profiler: Profiler = None,
              nthreads: int = None,
              iterator: _Union[str, MetaFunction] = None,
              extractor: _Union[str, MetaFunction] = None,
              mixer: _Union[str, MetaFunction] = None,
              mover: _Union[str, MetaFunction] = None) -> Populations:
    """Actually run the model... Real work happens here. The model
       will run until completion or until 'nsteps' have been
       completed, whichever happens first.

       Parameters
       ----------
       network: Network or Networks
            The network(s) on which to run the model
       infections: Infections
            The space used to record the infections
       rngs: list
            The list of random number generators to use, one per thread
       population: Population
            The initial population at the start of the model outbreak.
            This is also used to set the date and day of the start of
            the model outbreak
       seed: int
            The random number seed used for this model run. If this is
            None then a very random random number seed will be used
       output_dir: OutputFiles
            The directory to write all of the output into
       nsteps: int
            The maximum number of steps to run in the outbreak. If None
            then run until the outbreak has finished
       profiler: Profiler
            The profiler to use to profile - a new one is created if
            one isn't passed
       nthreads: int
            Number of threads over which to parallelise this model run
       iterator: MetaFunction or string
            Function that will be used to dynamically get the functions
            that will be used at each iteration to advance the
            model. Any additional files or parameters needed by these
            functions should be included in the `network.params` object.
       extractor: MetaFunction or string
            Function that will be used to dynamically get the functions
            that will be used at each iteration to extract data from
            the model run
       mixer: MetaFunction or string
            Function that will mix data from multiple demographics
            so that this is shared during a model run
       mover: MetaFunction or string
            Function that can move the population between different
            demographics

       Returns
       -------
       trajectory: Populations
            The trajectory of the population for every day of the model run
    """
    if iterator is None:
        from ..iterators._iterate_default import iterate_default
        iterator = iterate_default
    elif isinstance(iterator, str) or not accepts_stage(iterator):
        from ..iterators._iterate_custom import build_custom_iterator
        iterator = build_custom_iterator(iterator, __name__)

    if extractor is None:
        from ..extractors._extract_default import extract_default
        extractor = extract_default
    elif isinstance(extractor, str) or not accepts_stage(extractor):
        from ..extractors._extract_custom import build_custom_extractor
        extractor = build_custom_extractor(extractor, __name__)

    if mixer is None:
        from ..mixers._mix_default import mix_default
        mixer = mix_default
    elif isinstance(mixer, str) or not accepts_stage(mixer):
        from ..mixers._mix_custom import build_custom_mixer
        mixer = build_custom_mixer(mixer, __name__)

    if mover is None:
        from ..movers._move_default import move_default
        mover = move_default
    elif isinstance(mover, str) or not accepts_stage(mover):
        from ..movers._move_custom import build_custom_mover
        mover = build_custom_mover(mover, __name__)

    if profiler is None:
        from ._profiler import NullProfiler
        profiler = NullProfiler()

    p = profiler.start("run_model")

    params = network.params

    if params is None:
        return population

    from copy import deepcopy
    population = deepcopy(population)

    # create space to hold the population trajectory
    trajectory = Populations()

    p = p.start("clear_all_infections")
    infections.clear(nthreads=nthreads)
    p = p.stop()

    # create a workspace that is used as part of the "analyse" stage to
    # provide a scratch-pad while extracting data from the model
    workspace = Workspace.build(network=network)

    # get and call all of the functions that need to be called to
    # initialise the model run
    p = p.start("initialise_funcs")
    funcs = get_initialise_functions(network=network, population=population,
                                     infections=infections,
                                     output_dir=output_dir,
                                     workspace=workspace, rngs=rngs,
                                     iterator=iterator, extractor=extractor,
                                     mixer=mixer, mover=mover,
                                     nthreads=nthreads, profiler=p)

    for func in funcs:
        p = p.start(str(func))
        func(network=network, population=population,
             infections=infections, output_dir=output_dir,
             workspace=workspace, rngs=rngs, nthreads=nthreads,
             profiler=p)
        p = p.stop()

    p = p.stop()

    infecteds = population.infecteds

    # save the initial population
    trajectory.append(population)

    p = p.start("run_model_loop")
    iteration_count = 0

    from ._console import Console

    # keep looping until the outbreak is over or until we have completed
    # at least 5 loop iterations
    while (infecteds != 0) or (iteration_count < 5):
        # construct a new profiler of the same type as 'profiler'
        p2 = profiler.__class__()

        p2 = p2.start(f"timing for day {population.day}")

        Console.rule(f"Day {population.day}", style="iteration")

        start_population = population.population

        funcs = get_model_loop_functions(
            network=network, population=population,
            infections=infections,
            output_dir=output_dir,
            workspace=workspace, rngs=rngs,
            iterator=iterator, extractor=extractor,
            mixer=mixer, mover=mover,
            nthreads=nthreads, profiler=p)

        should_finish_early = False

        for func in funcs:
            p2 = p2.start(str(func))
            try:
                func(network=network, population=population,
                     infections=infections, output_dir=output_dir,
                     workspace=workspace, rngs=rngs, nthreads=nthreads,
                     profiler=p2)
            except StopIteration:
                # this function has signalled that the simulation
                # should now stop - we record this request but will
                # still let the other functions complete this
                # iteration
                Console.print(f"{func} has indicated that the model run "
                              f"should stop early. Will finish the run "
                              f"at the end of this iteration")
                should_finish_early = True

            p2 = p2.stop()

        if population.population != start_population:
            # something went wrong as the population should be conserved
            # during the day
            raise AssertionError(
                f"The total population changed during the day. This "
                f"should not happen and indicates a program bug. "
                f"The starting population was {start_population}, "
                f"while the end population is {population.population}. "
                f"Detail is {population}")

        infecteds = population.infecteds

        Console.print(f"Number of infections: {infecteds}")

        iteration_count += 1
        population.increment_day()

        p2 = p2.stop()

        if not p2.is_null():
            Console.print_profiler(p2)

        # save the population trajectory
        trajectory.append(population)

        if should_finish_early:
            Console.print(f"Exiting model run early due to function request")
            break

        elif nsteps is not None:
            if iteration_count >= nsteps:
                Console.print(
                    f"Exiting model run early as number of steps ({nsteps}) "
                    f"reached.")
                break

    # end of while loop
    p = p.stop()

    # finally get and call all of the functions needed to finalise
    # the model run, e.g. closing files, performing overall analyses,
    # writing summary files etc
    p = p.start("finalise_funcs")

    funcs = get_finalise_functions(network=network, population=population,
                                   infections=infections,
                                   output_dir=output_dir,
                                   workspace=workspace, rngs=rngs,
                                   iterator=iterator, extractor=extractor,
                                   mixer=mixer, mover=mover,
                                   nthreads=nthreads, trajectory=trajectory,
                                   profiler=p)

    for func in funcs:
        p = p.start(str(func))
        func(network=network, population=population,
             infections=infections, output_dir=output_dir,
             workspace=workspace, rngs=rngs, nthreads=nthreads,
             trajectory=trajectory, profiler=p)
        p = p.stop()

    p = p.stop()

    p.stop()

    if not p.is_null():
        Console.rule("Overall model timing")
        Console.print_profiler(p)

    Console.print(f"Infection died ... Ending on day {population.day}")

    # only send back the overall statistics
    return trajectory.strip_demographics()




__all__ = ["iterate_weekend"]


def iterate_weekend(nthreads: int = 1, **kwargs):
    """This returns the default list of 'advance_XXX' functions that
       are called in sequence for each weekend iteration of the model run.

       Parameters
       ----------
       nthreads: int
         The number of threads that will be used for each function.
         If this is 1, then the serial versions of the functions will
         be returned, else the parallel (OpenMP) versions will be
         returned

       Returns
       -------
       funcs: List[function]
         The list of functions that ```iterate``` will call in sequence
    """

    from ._advance_infprob import advance_infprob
    from ._advance_play import advance_play

    return [advance_infprob, advance_play]




from libc.stdio cimport FILE, fopen, fscanf, fclose, feof

from .._network import Network
from ._profiler import Profiler

from ._get_array_ptr cimport get_int_array_ptr, get_double_array_ptr

__all__ = ["assert_sane_network"]


def assert_sane_network(network: Network, profiler: Profiler):
    """This function runs through and checks that the passed network
       is sane. If it is not, it prints some information and raises
       an AssertionError
    """
    params = network.params
    nodes = network.nodes

    # read in the "work_size" file - this details the expected
    # population in each ward. This should match what has
    # been calculated via the links and play links files
    import os

    filename = params.input_files.work_size

    from ._console import Console

    if filename is None:
        Console.warning(f"Skipping network validation as work_size is null")

    elif not os.path.exists(filename):
        Console.warning(
                f"Skipping network validation as {filename} is unreadable")

    cdef char* fname
    cdef FILE* cfile

    p = profiler.start("read_and_validate")

    filename = filename.encode("UTF-8")
    fname = filename
    cfile = fopen(fname, "r")

    if cfile == NULL:
        raise FileNotFoundError(f"No such file or directory: {filename}")

    cdef int i = 0
    cdef nnodes = network.nnodes
    cdef double * nodes_denominator_n = get_double_array_ptr(
                                                    nodes.denominator_n)
    cdef double * nodes_denominator_p = get_double_array_ptr(
                                                    nodes.denominator_p)
    cdef double * nodes_play_suscept = get_double_array_ptr(
                                                    nodes.play_suscept)

    cdef int node_id = 0
    cdef double weight = 0.0
    cdef double node_weight = 0.0
    cdef int nfailed = 0

    while not feof(cfile):
        fscanf(cfile, "%d %lf\n", &node_id, &weight)

        if node_id < 0 or node_id > nnodes:
            Console.print(f"Corrupt index in work_size file: {node_id}")
            nfailed += 1
            continue

        node_weight = nodes_denominator_n[node_id]

        if weight != node_weight:
            Console.print(f"Incorrect weight for node {node_id}. Should be "
                          f"{weight}, but is instead {node_weight}.")
            nfailed += 1

        if nodes_denominator_p[node_id] != nodes_play_suscept[node_id]:
            Console.print(
                f"Disagreement in play_suscept for node {node_id}, "
                f"{nodes_denominator_p[node_id]} versus "
                f"{nodes_play_suscept[node_id]}")
            nfailed += 1

    if nfailed > 0:
        raise AssertionError(
            f"Invalid Network. Number of failures equals {nfailed}")

    p = p.stop()





from ._node import Node

__all__ = ["Nodes"]


class Nodes:
    """This is a container class for Nodes. This uses Numpy arrays
       to store a list of Node objects as a "struct of arrays".
       This should improve speed of loading and access.
    """

    def __init__(self, N: int = 0):
        """Create a container for up to "N" nodes"""
        if N <= 0:
            self._is_null = True
        else:
            self._is_null = False

            from .utils._array import create_double_array, create_int_array

            # Struct of arrays for each piece of data. See the
            # Node class for information about what each variable
            # holds. Code uses "-1" to represent a null value
            self.label = create_int_array(N, -1)

            self.begin_to = create_int_array(N, -1)
            self.end_to = create_int_array(N, -1)
            self.self_w = create_int_array(N, -1)

            self.begin_p = create_int_array(N, -1)
            self.end_p = create_int_array(N, -1)
            self.self_p = create_int_array(N, -1)

            self.day_foi = create_double_array(N, 0.0)
            self.night_foi = create_double_array(N, 0.0)

            self.play_suscept = create_double_array(N, 0.0)
            self.save_play_suscept = create_double_array(N, 0.0)

            self.denominator_n = create_double_array(N, 0.0)
            self.denominator_d = create_double_array(N, 0.0)

            self.denominator_p = create_double_array(N, 0.0)
            self.denominator_pd = create_double_array(N, 0.0)

            self.day_inf_prob = create_double_array(N, 0.0)
            self.night_inf_prob = create_double_array(N, 0.0)

            self.x = create_double_array(N, 0.0)  # no good initialiser for x
            self.y = create_double_array(N, 0.0)  # no good initialiser for y

    def is_null(self):
        return self._is_null

    def __len__(self):
        """Return the number of nodes in this container"""
        if self._is_null:
            return 0
        else:
            return len(self.label)

    def copy(self):
        """Return a copy of this network, using deep copies for things
           which are variable, but shallow copies for things which
           should stay the same (e.g. x and y). This copy assumes that
           you are not changing the actual links between nodes, e.g.
           it shallow-copies label, begin_to, end_to, self_w, begin_p, end_p
           and self_p. If you are changing the links then you will
           need to recreate these arrays.
        """
        from copy import copy, deepcopy

        nodes = copy(self)  # shallow copy

        # now deep copy the required elements
        nodes.day_foi = deepcopy(self.day_foi)
        nodes.night_foi = deepcopy(self.night_foi)

        nodes.play_suscept = deepcopy(self.play_suscept)
        nodes.save_play_suscept = deepcopy(self.save_play_suscept)

        nodes.denominator_n = deepcopy(self.denominator_n)
        nodes.denominator_d = deepcopy(self.denominator_d)

        nodes.denominator_p = deepcopy(self.denominator_p)
        nodes.denominator_pd = deepcopy(self.denominator_pd)

        nodes.day_inf_prob = deepcopy(self.day_inf_prob)
        nodes.night_inf_prob = deepcopy(self.night_inf_prob)

        return nodes

    def assert_not_null(self):
        """Assert that this collection of nodes is not null"""
        assert not self.is_null()

    def assert_valid_index(self, i):
        """Assert that the passed index 'i' is valid for this collection.
           This will return the index with  Python reverse indexing,
           so that "-1" refers to the last node in the collection
        """
        self.assert_not_null()
        n = len(self)

        if i < 0:
            i = n + i

        if i < 0 or i >= n:
            raise IndexError(f"Invalid node index {i}. Number of nodes "
                             f"in this container equals {n}")

        return i

    def population(self) -> int:
        """Return the total population of the nodes

           Returns
           -------
           int
               Total population
        """
        return sum(self.save_play_suscept)

    def __getitem__(self, i: int):
        """Return the node at index 'i'. Note that this is a
           deep copy of the node. Changing the node will not change
           the data in this container. To update the data in this
           container you need to use __setitem__, e.g. via the
           index operator (see __setitem__)
        """
        i = self.assert_valid_index(i)

        try:
            return Node(label=self.label[i],

                        begin_to=self.begin_to[i],
                        end_to=self.end_to[i],
                        self_w=self.self_w[i],

                        begin_p=self.begin_p[i],
                        end_p=self.end_p[i],
                        self_p=self.self_p[i],

                        day_foi=self.day_foi[i],
                        night_foi=self.night_foi[i],

                        play_suscept=self.play_suscept[i],
                        save_play_suscept=self.save_play_suscept[i],

                        denominator_n=self.denominator_n[i],
                        denominator_d=self.denominator_d[i],

                        denominator_p=self.denominator_p[i],
                        denominator_pd=self.denominator_pd[i],

                        x=self.x[i],
                        y=self.y[i])
        except Exception as e:
            from .utils._console import Console
            Console.error(f"Error at index {i}: "
                          f"size = {len(self)}: {e.__class__} {e}")
            raise e

    def __setitem__(self, i: int, value: Node):
        """Set the item as index 'i' equal to 'value'. This deep-copies
           'value' into this container
        """
        i = self.assert_valid_index(i)

        self.label[i] = value.label if value.label is not None else -1

        self.begin_to[i] = value.begin_to if value.begin_to is not None else -1
        self.end_to[i] = value.end_to if value.end_to is not None else -1
        self.self_w[i] = value.self_w if value.self_w is not None else -1

        self.begin_p[i] = value.begin_p if value.begin_p is not None else -1
        self.end_p[i] = value.end_p if value.end_p is not None else -1
        self.self_p[i] = value.self._p if value.self_p is not None else -1

        self.day_foi[i] = value.day_foi if value.day_foi is not None else 0.0
        self.night_foi[i] = value.night_foi if value.night_foi is not None else 0.0

        self.play_suscept[i] = value.play_suscept if value.play_suscept is not None else 0.0
        self.save_play_suscept[i] = value.save_play_suscept if value.save_play_suscept is not None else 0.0

        self.denominator_n[i] = value.denominator_n if value.denominator_n is not None else 0.0
        self.denominator_d[i] = value.denominator_d if value.denominator_d is not None else 0.0

        self.denominator_p[i] = value.denominator_p if value.denominator_p is not None else 0.0
        self.denominator_pd[i] = value.denominator_pd if value.denominator_pd is not None else 0.0

        self.x[i] = value.x if value.x is not None else 0.0
        self.y[i] = value.y if value.y is not None else 0.0

    def resize(self, N: int):
        """Resize this container to hold 'N' nodes. This will expand
           the container if 'N' is greater than len(self), or will
           contract the container (deleting excess nodes) if 'N'
           is less than len(self). This function is called typically
           when you pre-allocate a large Nodes container, and then
           want to reduce the size to fit the number of loaded nodes
        """
        if self.is_null():
            # Assign to a new Nodes object created with this size
            self.__dict__ = Nodes(N).__dict__
            return
        elif N <= 0:
            self.__dict__ = Nodes(0).__dict__

        size = len(self)

        if N == size:
            return

        from .utils._array import resize_array
        self.label = resize_array(self.label, N, -1)

        self.begin_to = resize_array(self.begin_to, N, -1)
        self.end_to = resize_array(self.end_to, N, -1)
        self.self_w = resize_array(self.self_w, N, -1)

        self.begin_p = resize_array(self.begin_p, N, -1)
        self.end_p = resize_array(self.end_p, N, -1)
        self.self_p = resize_array(self.self_p, N, -1)

        self.day_foi = resize_array(self.day_foi, N, 0.0)
        self.night_foi = resize_array(self.night_foi, N, 0.0)

        self.play_suscept = resize_array(self.play_suscept, N, 0.0)
        self.save_play_suscept = resize_array(self.save_play_suscept, N, 0.0)

        self.denominator_n = resize_array(self.denominator_n, N, 0.0)
        self.denominator_d = resize_array(self.denominator_d, N, 0.0)

        self.denominator_p = resize_array(self.denominator_p, N, 0.0)
        self.denominator_pd = resize_array(self.denominator_pd, N, 0.0)

        self.x = resize_array(self.x, N, 0.0)
        self.y = resize_array(self.y, N, 0.0)

    def scale_susceptibles(self, ratio: any = None,
                           work_ratio: any = None, play_ratio: any = None):
        """Scale the number of susceptibles in these Nodes
           by the passed scale ratios. These can be values, e.g.
           ratio = 2.0 will scale the total number of susceptibles
           in each ward by 2.0. They can also be lists of values,
           where ward[i] will be scaled by ratio[i]. They can also
           be dictionaries, e.g. ward[i] scaled by ratio[i]

           Parameters
           ----------
           ratio: None, float, list or dict
             The amount by which to scale the total population of
             susceptibles - evenly scales the work and play populations
           work_ratio: None, float, list or dict
             Scale only the work population of susceptibles
           play_ratio: None, float, list or dict
             Scale only the play population of susceptibles

           Returns
           -------
           None
        """
        from .utils._scale_susceptibles import scale_node_susceptibles
        scale_node_susceptibles(nodes=self, ratio=ratio,
                                work_ratio=work_ratio,
                                play_ratio=play_ratio)



#!/bin/env/python3
#cython: linetrace=False
# MUST ALWAYS DISABLE AS WAY TOO SLOW FOR ITERATE

cimport cython
from cython.parallel import parallel, prange

from .._network import Network

from ..utils._profiler import Profiler
from ..utils._get_functions import call_function_on_network

from ..utils._rate_to_prob cimport rate_to_prob

from ..utils._get_array_ptr cimport get_int_array_ptr, get_double_array_ptr

__all__ = ["advance_infprob", "advance_infprob_omp",
           "advance_infprob_serial"]


def advance_infprob_omp(network: Network, nthreads: int,
                        profiler: Profiler,
                        scale_rate: float = 1.0,
                        **kwargs):
    """Advance the calculation of the day and night infection probabilities
       for each ward. You need to call this function after you have
       changed them, and before you use them in, e.g. advance_fixed
       and advance_play. This is the parallel version of this function

       Parameters
       ----------
       network: Network
         The network being modelled
       nthreads: int
         The number of threads over which to parallelise the calculation
       profiler: Profiler
         The profiler used to profile this calculation
       scale_rate: float
         Optional parameter to scale the calculated infection rates.
         This can very bluntly model the impact of broad measures
         to reduce the infection rate
       kwargs:
         Extra arguments that may be used by other advancers, but which
         are not used by advance_infprob
    """
    wards = network.nodes
    params = network.params

    # Copy arguments from Python into C cdef variables
    cdef double length_day = params.length_day
    cdef double sclrate = scale_rate

    if sclrate < 0:
        sclrate = 0.0

    cdef double * wards_day_foi = get_double_array_ptr(wards.day_foi)
    cdef double * wards_night_foi = get_double_array_ptr(wards.night_foi)

    cdef double * wards_denominator_d = get_double_array_ptr(
                                                    wards.denominator_d)
    cdef double * wards_denominator_n = get_double_array_ptr(
                                                    wards.denominator_n)
    cdef double * wards_denominator_p = get_double_array_ptr(
                                                    wards.denominator_p)
    cdef double * wards_denominator_pd = get_double_array_ptr(
                                                    wards.denominator_pd)

    cdef double * wards_day_inf_prob = get_double_array_ptr(
                                                    wards.day_inf_prob)
    cdef double * wards_night_inf_prob = get_double_array_ptr(
                                                    wards.night_inf_prob)

    # create and initialise variables used in the loop
    cdef double rate = 0.0
    cdef double inf_prob = 0.0
    cdef double denom = 0.0

    cdef int num_threads = nthreads
    cdef int thread_id = 0
    cdef int j = 0
    cdef int nnodes_plus_one = network.nnodes + 1

    ## Finally(!) we can now declare the actual loop.
    ## This loops in parallel over all wards to calculate
    ## the day and night infection probabilities for
    ## all wards

    p = profiler.start("infprob")
    with nogil, parallel(num_threads=num_threads):
        for j in prange(1, nnodes_plus_one, schedule="static"):
            # pre-calculate the day and night infection probability
            # for each ward
            denom = wards_denominator_d[j] + wards_denominator_pd[j]

            if denom != 0.0:
                rate = (length_day * wards_day_foi[j]) / denom
                wards_day_inf_prob[j] = rate_to_prob(sclrate * rate)
            else:
                wards_day_inf_prob[j] = 0.0

            denom = wards_denominator_n[j] + wards_denominator_p[j]

            if denom != 0.0:
                rate = (1.0 - length_day) * (wards_night_foi[j]) / denom
                wards_night_inf_prob[j] = rate_to_prob(sclrate * rate)
            else:
                wards_night_inf_prob[j] = 0.0
        # end of loop over wards
    # end of parallel
    p = p.stop()


def advance_infprob_serial(network: Network,
                           profiler: Profiler,
                           scale_rate: float = 1.0,
                           **kwargs):
    """Advance the calculation of the day and night infection probabilities
       for each ward. You need to call this function after you have
       changed them, and before you use them in, e.g. advance_fixed
       and advance_play. This is the serial version of this function

       Parameters
       ----------
       network: Network
         The network being modelled
       profiler: Profiler
         The profiler used to profile this calculation
       kwargs:
         Extra arguments that may be used by other advancers, but which
         are not used by advance_infprob
    """

    wards = network.nodes
    params = network.params

    # Copy arguments from Python into C cdef variables
    cdef double length_day = params.length_day
    cdef double sclrate = scale_rate

    if sclrate < 0:
        sclrate = 0.0

    cdef double * wards_day_foi = get_double_array_ptr(wards.day_foi)
    cdef double * wards_night_foi = get_double_array_ptr(wards.night_foi)

    cdef double * wards_denominator_d = get_double_array_ptr(
                                                    wards.denominator_d)
    cdef double * wards_denominator_n = get_double_array_ptr(
                                                    wards.denominator_n)
    cdef double * wards_denominator_p = get_double_array_ptr(
                                                    wards.denominator_p)
    cdef double * wards_denominator_pd = get_double_array_ptr(
                                                    wards.denominator_pd)

    cdef double * wards_day_inf_prob = get_double_array_ptr(
                                                    wards.day_inf_prob)
    cdef double * wards_night_inf_prob = get_double_array_ptr(
                                                    wards.night_inf_prob)

    # create and initialise variables used in the loop
    cdef double rate = 0.0
    cdef double inf_prob = 0.0
    cdef double denom = 0.0

    cdef int j = 0
    cdef int nnodes_plus_one = network.nnodes + 1

    ## Finally(!) we can now declare the actual loop.
    ## This loops in parallel over all wards to calculate
    ## the day and night infection probabilities for
    ## all wards

    p = profiler.start("infprob")
    with nogil:
        for j in range(1, nnodes_plus_one):
            # pre-calculate the day and night infection probability
            # for each ward
            denom = wards_denominator_d[j] + wards_denominator_pd[j]

            if denom != 0.0:
                rate = (length_day * wards_day_foi[j]) / denom
                wards_day_inf_prob[j] = rate_to_prob(sclrate * rate)
            else:
                wards_day_inf_prob[j] = 0.0

            denom = wards_denominator_n[j] + wards_denominator_p[j]

            if denom != 0.0:
                rate = (1.0 - length_day) * (wards_night_foi[j]) / denom
                wards_night_inf_prob[j] = rate_to_prob(sclrate * rate)
            else:
                wards_night_inf_prob[j] = 0.0
        # end of loop over wards
    # end of parallel
    p = p.stop()


def advance_infprob(nthreads: int, **kwargs):
    """Advance the calculation of the day and night infection probabilities
       for each ward. You need to call this function after you have
       changed them, and before you use them in, e.g. advance_fixed
       and advance_play. This is the parallel version of this function

       Parameters
       ----------
       network: Network
         The network being modelled
       nthreads: int
         The number of threads over which to parallelise the calculation
       profiler: Profiler
         The profiler used to profile this calculation
       scale_rate: float
         Optional parameter to scale the calculated infection rates.
         This can very bluntly model the impact of broad measures
         to reduce the infection rate
       kwargs:
         Extra arguments that may be used by other advancers, but which
         are not used by advance_infprob
    """
    call_function_on_network(nthreads=nthreads,
                             func=advance_infprob_serial,
                             parallel=advance_infprob_omp,
                             **kwargs)




from typing import List as _List

__all__ = ["align_strings"]


def align_strings(strings: _List[str], sep: str = ":") -> _List[str]:
    """Align the passed list of strings such that they are aligned
       against every appearance of 'sep'
    """
    results = []

    size = []

    for string in strings:
        words = string.split(sep)

        for i, word in enumerate(words):
            if i >= len(size):
                size.append(len(word))
            elif len(word) > size[i]:
                size[i] = len(word)

    for string in strings:
        words = string.split(sep)

        result = ""

        for i, word in enumerate(words):
            diff = size[i] - len(word)
            result += diff * " " + word

            if i < len(words) - 1:
                result += sep

        results.append(result)

    return results




from dataclasses import dataclass as _dataclass
from typing import List as _List
from copy import deepcopy as _deepcopy
from datetime import date as _date

__all__ = ["Population", "Populations"]


@_dataclass
class Population:
    """This class holds information about the progress of the
       disease through the population
    """
    #: The initial population loaded into the model
    initial: int = 0
    #: The number of members who could be infected
    susceptibles: int = 0
    #: The number of latent infections
    latent: int = 0
    #: The total number of infections
    total: int = 0
    #: The total number who are removed from the outbreak,
    #: either because they have recovered, or are otherwise
    #: no longer able to be infected
    recovereds: int = 0
    #: The number infected in all wards
    n_inf_wards: int = 0
    #: The scale_uv parameter that can be used to affect the
    #: foi calculation. A value of 1.0 means do nothing
    scale_uv: float = 1.0
    #: The day in the outbreak of this record (e.g. day 0, day 10 etc.)
    day: int = 0
    #: The date in the outbreak of this record
    date: _date = None

    #: The populations in each of the multi-demographic subnetworks
    subpops = None

    @property
    def population(self) -> int:
        """The total population in all wards"""
        return self.susceptibles + self.latent + self.total + self.recovereds

    @property
    def infecteds(self) -> int:
        """The number who are infected across all wards"""
        return self.total + self.latent

    def increment_day(self, ndays: int = 1) -> None:
        """Advance the day count by 'ndays' (default 1)"""
        self.day += ndays

        if self.subpops is not None:
            for pop in self.subpops:
                pop.day += ndays

        if self.date:
            from datetime import timedelta
            self.date += timedelta(days=1)

            if self.subpops is not None:
                for pop in self.subpops:
                    pop.date = self.date

    def specialise(self, network):
        """Specialise this population for the passed Networks"""
        subpops = []

        from copy import deepcopy

        self.subpops = None

        for i in range(0, len(network.subnets)):
            subpops.append(deepcopy(self))

        self.subpops = subpops

    def __str__(self):
        s = f"DAY: {self.day} " \
            f"S: {self.susceptibles}    " \
            f"E: {self.latent}    " \
            f"I: {self.total}    " \
            f"R: {self.recovereds}    " \
            f"IW: {self.n_inf_wards}   " \
            f"UV: {self.scale_uv}   " \
            f"TOTAL POPULATION {self.population}"

        if self.date:
            return f"{self.date.isoformat()}: {s}"
        else:
            return s

    def assert_sane(self):
        """Assert that this population is sane, i.e. the totals within
           this population and with the sub-populations all add up to
           the correct values
        """
        errors = []

        t = self.susceptibles + self.latent + self.total + self.recovereds

        if t != self.population:
            errors.append(f"Disagreement in total overall population: "
                          f"{t} versus {self.population}")

        if self.subpops is not None and len(self.subpops) > 0:
            S = 0
            E = 0
            I = 0
            R = 0
            P = 0

            for subpop in self.subpops:
                S += subpop.susceptibles
                E += subpop.latent
                I += subpop.infecteds
                R += subpop.recovereds
                P += subpop.population

            if S != self.susceptibles:
                errors.append(f"Disagreement in S: {S} "
                              f"versus {self.susceptibles}")

            if E != self.latent:
                errors.append(f"Disagreement in E: {E} "
                              f"versus {self.latent}")

            if I != self.infecteds:
                errors.append(f"Disagreement in I: {I} "
                              f"versus {self.infecteds}")

            if R != self.recovereds:
                errors.append(f"Disagreement in R: {R} "
                              f"versus {self.recovereds}")

            if P != self.population:
                errors.append(f"Disagreement in Population: {P} "
                              f"versus {self.population}")

        if len(errors) > 0:
            errors = "\nERROR: ".join(errors)
            from .utils._console import Console
            Console.error(errors)
            raise AssertionError(f"Disagreement in population sums!")

    def summary(self, demographics=None):
        """Return a short summary string that is suitable to be printed
           out during a model run

           Returns
           -------
           summary: str
             The short summary string
        """
        summary = f"S: {self.susceptibles}  E: {self.latent}  " \
                  f"I: {self.total}  R: {self.recovereds}  " \
                  f"IW: {self.n_inf_wards}  POPULATION: {self.population}"

        if self.subpops is None or len(self.subpops) == 0:
            return summary

        subs = []
        for i, subpop in enumerate(self.subpops):
            if demographics is not None:
                name = demographics.get_name(i)
                subs.append(f"{name}  {subpop.summary()}")
            else:
                subs.append(f"{i}  {subpop.summary()}")

        from .utils._align_strings import align_strings
        subs = align_strings(subs, ":")

        return f"{summary}\n  " + "\n  ".join(subs)


@_dataclass
class Populations:
    """This class holds the trajectory of Population objects recorded
       for every step (day) of a model outbreak
    """
    #: The trajectory of Population objects
    _trajectory: _List[Population] = None

    def __str__(self):
        if len(self) == 0:
            return "Populations:empty"
        else:
            return f"Latest: {self._trajectory[-1]}"

    def __getitem__(self, i: int):
        """Return the ith Population in the trajectory"""
        if self._trajectory is None:
            raise IndexError("No trajectory data collected")
        else:
            return self._trajectory[i]

    def __len__(self):
        if self._trajectory is None:
            return 0
        else:
            return len(self._trajectory)

    def strip_demographics(self):
        """Remove the demographics information from this trajectory. This
           makes it much smaller and easier to transmit over a network
        """
        for value in self._trajectory:
            value._subpops = None

        return self

    def append(self, population: Population):
        """Append the next step in the trajectory.

           Parameters
           ----------
           population: Population
             The population to append to this list
        """
        if not isinstance(population, Population):
            raise TypeError("Only Population objects should be recorded!")

        if self._trajectory is None:
            self._trajectory = []

        self._trajectory.append(_deepcopy(population))




from dataclasses import dataclass as _dataclass
from typing import List as _List
from typing import Union as _Union

from ._network import Network
from ._networks import Networks

__all__ = ["Workspace"]


@_dataclass
class Workspace:
    """This class provides a workspace for the running calculation.
       This pre-allocates all of the memory into arrays, which
       can then be used via cython memory views
    """
    #: Number of disease classes (stages)
    n_inf_classes: int = 0

    #: Number of wards (nodes)
    nnodes: int = 0

    #: Size of population in each disease stage for work infections
    inf_tot: _List[int] = None
    #: Size of population in each disease stage for play infections
    pinf_tot: _List[int] = None
    #: Number of wards with at least one individual in this disease stage
    n_inf_wards: _List[int] = None

    #: Total number of infections in each ward over the last day
    #: This is also equal to the prevalence
    total_inf_ward: _List[int] = None

    #: Number of new infections in each ward over the last day
    total_new_inf_ward: _List[int] = None

    #: The incidence of the infection (sum of infections up to
    #: disease_class == 2)
    incidence: _List[int] = None

    #: The size of the S population in each ward
    S_in_wards: _List[int] = None

    #: The size of the E population in each ward
    E_in_wards: _List[int] = None

    #: The size of the I population in each ward
    I_in_wards: _List[int] = None

    #: The size of the R population in each ward
    R_in_wards: _List[int] = None

    #: The sub-workspaces used for the subnets of a
    #: multi-demographic Networks (list[Workspace])
    subspaces = None

    @staticmethod
    def build(network: _Union[Network, Networks]):
        """Create the workspace needed to run the model for the
           passed network
        """
        params = network.params

        workspace = Workspace()

        if isinstance(network, Network):
            n_inf_classes = params.disease_params.N_INF_CLASSES()

            workspace.n_inf_classes = params.disease_params.N_INF_CLASSES()
            workspace.nnodes = network.nnodes

            size = workspace.nnodes + 1  # 1-indexed

            from .utils._array import create_int_array

            workspace.inf_tot = create_int_array(n_inf_classes, 0)
            workspace.pinf_tot = create_int_array(n_inf_classes, 0)
            workspace.n_inf_wards = create_int_array(n_inf_classes, 0)

            workspace.total_inf_ward = create_int_array(size, 0)
            workspace.total_new_inf_ward = create_int_array(size, 0)
            workspace.incidence = create_int_array(size, 0)

            workspace.S_in_wards = create_int_array(size, 0)
            workspace.E_in_wards = create_int_array(size, 0)
            workspace.I_in_wards = create_int_array(size, 0)
            workspace.R_in_wards = create_int_array(size, 0)

        elif isinstance(network, Networks):
            workspace = Workspace.build(network.overall)

            subspaces = []

            for subnet in network.subnets:
                subspaces.append(Workspace.build(subnet))

            workspace.subspaces = subspaces

        return workspace

    def zero_all(self, zero_subspaces=True):
        """Reset the values of all of the arrays to zero.
           By default we zero the subspace networks
           (change this by setting zero_subspaces to False)
        """
        for i in range(0, self.n_inf_classes):
            self.inf_tot[i] = 0
            self.pinf_tot[i] = 0
            self.n_inf_wards[i] = 0

        for i in range(0, self.nnodes+1):
            self.total_inf_ward[i] = 0
            self.total_new_inf_ward[i] = 0
            self.incidence[i] = 0
            self.S_in_wards[i] = 0
            self.E_in_wards[i] = 0
            self.I_in_wards[i] = 0
            self.R_in_wards[i] = 0

        if zero_subspaces and self.subspaces is not None:
            for subspace in self.subspaces:
                subspace.zero_all()




cdef double * get_double_array_ptr(double_array):
    """Return the raw C pointer to the passed double array which was
       created using create_double_array
    """
    cdef double [::1] a = double_array
    return &(a[0])


cdef int * get_int_array_ptr(int_array):
    """Return the raw C pointer to the passed int array which was
       created using create_int_array
    """
    cdef int [::1] a = int_array
    return &(a[0])




from typing import Union as _Union
from typing import Dict as _Dict

from .._network import Network
from .._networks import Networks
from .._demographics import Demographics
from .._parameters import Parameters
from .._outputfiles import OutputFiles

import os
import sys

__all__ = ["run_worker", "prepare_worker"]

global_network = None


def prepare_worker(params: Parameters, demographics: Demographics,
                   options: _Dict[str, any]) -> _Union[Network, Networks]:
    """Prepare a worker to receive work to run a model using the passed
       parameters. This will build the network specified by the
       parameters and will store it in global memory ready to
       be used for a model run. Note that these are
       silent, printing nothing to stdout or stderr

       Parameters
       ----------
       params: Parameters
         Parameters used to build the network
       demographics: Demographics
         If not None, then demographics used to specialise the Network
         into Networks
    """
    # switch off printing to stdout and stderr
    global global_network

    max_nodes = options["max_nodes"]
    max_links = options["max_links"]
    nthreads = options["nthreads"]

    del options["max_nodes"]
    del options["max_links"]

    profiler = options["profiler"]

    if global_network is None:
        network = Network.build(params=params,
                                profiler=profiler,
                                max_nodes=max_nodes,
                                max_links=max_links)

        if demographics is not None:
            network = demographics.specialise(network,
                                              nthreads=nthreads,
                                              profiler=profiler)

        global_network = network

    # always work in a copy
    network = global_network.copy()
    network.update(params=params, demographics=demographics,
                   nthreads=nthreads, profiler=profiler)

    return network


def run_worker(arguments):
    """Ask the worker to run a model using the passed variables and
       options. This will write to options['output_dir'] and will
       also return the population object that contains the final
       population data.

       WARNING - the iterator and extractor arguments rely on the
       workers starting in the same directory as the main process,
       so that they can load the same python files (if the user
       is using a custom iterator or extractor)
    """
    params = arguments["params"]
    demographics = arguments["demographics"]
    options = arguments["options"]

    # next, run the job, writing to output
    outdir = options["output_dir"]
    auto_bzip = options["auto_bzip"]
    del options["auto_bzip"]

    from ._console import Console

    with OutputFiles(outdir, check_empty=False, force_empty=False,
                     prompt=None, auto_bzip=auto_bzip) as output_dir:
        with Console.redirect_output(outdir=outdir, auto_bzip=auto_bzip):
            # first, build and prepare the Network(s). This is built once
            # from the parameters and demographics by loading files from
            # the filesystem, as sending this over the physical network
            # would be too expensive. Subsequent calls to this function
            # after the Network(s) has been built will call
            # network.update(params, demographics)
            network = prepare_worker(params=params, demographics=demographics,
                                     options=options)

            # if the user wanted to remove this directory then they would
            # have done so in the main process - no need to check again
            options["output_dir"] = output_dir

            output = network.run(**options)

            return output




__all__ = ["mix_default"]


def mix_default(**kwargs):
    """This is the default mixer. By default, nothing extra is mixed
       at any stage of the model run
    """
    from ._mix_none import mix_none
    return mix_none()
